#!python
#cython: language_level=3


# Original code developed by J. Derek Tucker.
# 03/25/2019: Modified by Pablo Marcos <pablo.marcosm@estudiante.uam.es>.


import numpy as np
from numpy.linalg import norm

cimport numpy as np
from cpython cimport array

cdef extern from "dp_grid.h":
    void dp_optimum_reparam(double *Q1, double *T1, double *Q2, double *T2,
                            int m1, int n1, int n2, double *tv1, double *tv2,
                            int n1v, int n2v, double *G, double *T,
                            double *size, double lam1, int nbhd_dim)



def coptimum_reparam_n(np.ndarray[double, ndim=1, mode="c"] mq,
                      np.ndarray[double, ndim=1, mode="c"] time,
                      np.ndarray[double, ndim=2, mode="c"] q,
                      lam1=0.0, nbhd_dim=7):
    """
    cython interface calculates the warping to align a set of srfs q to a
    single srsf mq.

    :param mq: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of N srsfs with M samples
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of
        q with mq

    """
    cdef int M, N, n1
    cdef double lam
    mq = mq / norm(mq)
    M, N = q.shape[0], q.shape[1]
    n1 = 1
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] qi = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    gam = np.zeros((M, N))
    sizes = np.zeros(N, dtype=np.int32)
    Go = np.zeros((M, N))
    To = np.zeros((M, N))
    for k in range(0, N):
        qi = q[:, k] / norm(q[:, k])
        qi = np.ascontiguousarray(qi)

        dp_optimum_reparam(&mq[0], &time[0], &qi[0], &time[0], n1, M, M,
                                  &time[0], &time[0], M, M, &G[0], &T[0],
                                  &size[0], lam, nbhd_dim)
        sizes[k] = np.int32(size)
        Go[:, k] = G
        To[:, k] = T

    for k in range(0, N):
        gam0 = np.interp(time, To[0:sizes[k], k], Go[0:sizes[k], k])
        gam[:, k] = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparam_n2(np.ndarray[double, ndim=2, mode="c"] q1,
                       np.ndarray[double, ndim=1, mode="c"] time,
                       np.ndarray[double, ndim=2, mode="c"] q2,
                       lam1=0.0, nbhd_dim=7):
    """
    cython interface calculates the warping to align a set of srsfs q1 to
    another set of srsfs q2

    :param q1: numpy ndarray of shape (M,N) of M srsfs with N samples
    :param time: vector of size N describing the sample points
    :param q2: numpy ndarray of shape (M,N) of M srsfs with N samples
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of
        q with mq

    """
    cdef int M, N, n1
    cdef double lam

    M, N = q1.shape[0], q1.shape[1]
    n1 = 1
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    gam = np.zeros((M, N))
    sizes = np.zeros(N, dtype=np.int32)
    Go = np.zeros((M, N))
    To = np.zeros((M, N))
    for k in range(0, N):
        q1i = q1[:, k] / norm(q1[:, k])
        q2i = q2[:, k] / norm(q2[:, k])
        q1i = np.ascontiguousarray(q1i)
        q2i = np.ascontiguousarray(q2i)

        dp_optimum_reparam(&q1i[0], &time[0], &q2i[0], &time[0], n1,
                                  M, M, &time[0], &time[0], M, M, &G[0],
                                  &T[0], &size[0], lam, nbhd_dim)
        sizes[k] = np.int32(size)
        Go[:, k] = G
        To[:, k] = T

    for k in range(0, N):
        gam0 = np.interp(time, To[0:sizes[k], k], Go[0:sizes[k], k])
        gam[:, k] = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparam(np.ndarray[double, ndim=1, mode="c"] q1,
                     np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=1, mode="c"] q2,
                     lam1=0.0, nbhd_dim=7):
    """
    cython interface for calculates the warping to align srsf q2 to q1

    :param q1: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q2: vector of size N samples of second SRSF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, n1
    cdef double lam
    M = q1.shape[0]
    n1 = 1
    lam = lam1
    q1 = q1 / norm(q1)
    q2 = q2 / norm(q2)
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    sizes = np.zeros(1, dtype=np.int32)
    Go = np.zeros((M, 1))
    To = np.zeros((M, 1))
    dp_optimum_reparam(&q1[0], &time[0], &q2[0], &time[0], n1, M, M,
                              &time[0], &time[0], M, M, &G[0],
                              &T[0], &size[0], lam, nbhd_dim)
    sizes = np.int32(size)
    Go[:, 0] = G
    To[:, 0] = T
    gam0 = np.interp(time, To[0:sizes[0], 0], Go[0:sizes[0], 0])
    gam = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam



"""Registration of functional data module.

This module contains routines related to the registration procedure.
"""
import collections

import numpy
import scipy.integrate
from scipy.interpolate import PchipInterpolator

__author__ = "Pablo Marcos Manchón"
__email__ = "pablo.marcosm@estudiante.uam.es"


def mse_decomposition(original_fdata, registered_fdata, warping_function=None,
                      *, eval_points=None):
    r"""Compute mean square error measures for amplitude and phase variation.

    Once the registration has taken place, this function computes two mean
    squared error measures, one for amplitude variation, and the other for
    phase variation. It also computes a squared multiple correlation index
    of the amount of variation in the unregistered functions is due to phase.

    Let :math:`x_i(t),y_i(t)` be the unregistered and registered functions
    respectively. The total mean square error measure (see [RGS09-8-5]_) is
    defined as


    .. math::
        \text{MSE}_{total}=
        \frac{1}{N}\sum_{i=1}^{N}\int[x_i(t)-\overline x(t)]^2dt

    We define the constant :math:`C_R` as

    .. math::

        C_R = 1 + \frac{\frac{1}{N}\sum_{i}^{N}\int [Dh_i(t)-\overline{Dh}(t)]
        [ y_i^2(t)- \overline{y^2}(t) ]dt}
        {\frac{1}{N} \sum_{i}^{N} \int y_i^2(t)dt}

    Whose structure is related to the covariation between the deformation
    functions :math:`Dh_i(t)` and the squared registered functions
    :math:`y_i^2(t)`. When these two sets of functions are independents
    :math:`C_R=1`, as in the case of shift registration.

    The measures of amplitude and phase mean square error are

    .. math::
        \text{MSE}_{amp} =  C_R \frac{1}{N}
        \sum_{i=1}^{N} \int \left [ y_i(t) - \overline{y}(t) \right ]^2 dt

    .. math::
        \text{MSE}_{phase}=
        \int \left [C_R \overline{y}^2(t) - \overline{x}^2(t) \right]dt

    It can be shown that

    .. math::
        \text{MSE}_{total} = \text{MSE}_{amp} + \text{MSE}_{phase}

    The squared multiple correlation index of the proportion of the total
    variation due to phase is defined as:

    .. math::
        R^2 = \frac{\text{MSE}_{phase}}{\text{MSE}_{total}}

    See [KR08-3]_ for a detailed explanation.


    Args:
        original_fdata (:class:`FData`): Unregistered functions.
        regfd (:class:`FData`): Registered functions.
        warping_function (:class:`FData`): Warping functions.
        eval_points: (array_like, optional): Set of points where the
            functions are evaluated to obtain a discrete representation.


    Returns:
        :class:`collections.namedtuple`: Tuple with amplitude mean square error
        :math:`\text{MSE}_{amp}`, phase mean square error
        :math:`\text{MSE}_{phase}`, squared correlation index :math:`R^2`
        and constant :math:`C_R`.

    Raises:
        ValueError: If the curves do not have the same number of samples.

    References:
        ..  [KR08-3] Kneip, Alois & Ramsay, James. (2008).  Quantifying
            amplitude and phase variation. In *Combining Registration and
            Fitting for Functional Models* (pp. 14-15). Journal of the American
            Statistical Association.
        ..  [RGS09-8-5] Ramsay J.O., Giles Hooker & Spencer Graves (2009). In
            *Functional Data Analysis with R and Matlab* (pp. 125-126).
            Springer.

    Examples:

        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (landmark_registration_warping,
        ...                               mse_decomposition)


        We will create and register data.

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()
        >>> warping = landmark_registration_warping(fd, landmarks)
        >>> fd_registered = fd.compose(warping)
        >>> mse_amp, mse_pha, rsq, cr = mse_decomposition(fd, fd_registered, warping)

        Mean square error produced by the amplitude variation.

        >>> f'{mse_amp:.6f}'
        '0.000987'

        In this example we can observe that the main part of the mean square
        error is due to the phase variation.

        >>> f'{mse_pha:.6f}'
        '0.115769'

        Nearly 99% of the variation is due to phase.

        >>> f'{rsq:.6f}'
        '0.991549'

    """

    if registered_fdata.ndim_domain != 1 or registered_fdata.ndim_image != 1:
        raise NotImplementedError

    if original_fdata.nsamples != registered_fdata.nsamples:
        raise ValueError(f"the registered and unregistered curves must have "
                         f"the same number of samples "
                         f"({registered_fdata.nsamples})!= "
                         f"({original_fdata.nsamples})")

    if warping_function is not None and (warping_function.nsamples
                                         != original_fdata.nsamples):
        raise ValueError(f"the registered curves and the warping functions must"
                         f" have the same number of samples "
                         f"({registered_fdata.nsamples})"
                         f"!=({warping_function.nsamples})")

    # Creates the mesh to discretize the functions
    if eval_points is None:
        try:
            eval_points = registered_fdata.sample_points[0]

        except AttributeError:
            nfine = max(registered_fdata.basis.nbasis * 10 + 1, 201)
            domain_range = registered_fdata.domain_range[0]
            eval_points = numpy.linspace(*domain_range, nfine)
    else:
        eval_points = numpy.asarray(eval_points)

    x_fine = original_fdata.evaluate(eval_points, keepdims=False)
    y_fine = registered_fdata.evaluate(eval_points, keepdims=False)
    mu_fine = x_fine.mean(axis=0)  # Mean unregistered function
    eta_fine = y_fine.mean(axis=0)  # Mean registered function
    mu_fine_sq = numpy.square(mu_fine)
    eta_fine_sq = numpy.square(eta_fine)

    # Total mean square error of the original funtions
    # mse_total = scipy.integrate.simps(
    #    numpy.mean(numpy.square(x_fine - mu_fine), axis=0),
    #    eval_points)

    cr = 1.  # Constant related to the covariation between the deformation
    # functions and y^2

    # If the warping functions are not provided, are suppose to be independent
    if warping_function is not None:
        # Derivates warping functions
        dh_fine = warping_function.evaluate(eval_points, derivative=1,
                                            keepdims=False)
        dh_fine_mean = dh_fine.mean(axis=0)
        dh_fine_center = dh_fine - dh_fine_mean

        y_fine_sq = numpy.square(y_fine)  # y^2
        y_fine_sq_center = numpy.subtract(
            y_fine_sq, eta_fine_sq)  # y^2 - E[y^2]

        covariate = numpy.inner(dh_fine_center.T, y_fine_sq_center.T)
        covariate = covariate.mean(axis=0)
        cr += numpy.divide(scipy.integrate.simps(covariate,
                                                 eval_points),
                           scipy.integrate.simps(eta_fine_sq,
                                                 eval_points))

    # mse due to phase variation
    mse_pha = scipy.integrate.simps(cr*eta_fine_sq - mu_fine_sq, eval_points)

    # mse due to amplitude variation
    # mse_amp = mse_total - mse_pha
    y_fine_center = numpy.subtract(y_fine, eta_fine)
    y_fine_center_sq = numpy.square(y_fine_center, out=y_fine_center)
    y_fine_center_sq_mean = y_fine_center_sq.mean(axis=0)

    mse_amp = scipy.integrate.simps(y_fine_center_sq_mean, eval_points)

    # Total mean square error of the original funtions
    mse_total = mse_pha + mse_amp

    # squared correlation measure of proportion of phase variation
    rsq = mse_pha / (mse_total)

    mse_decomp = collections.namedtuple('mse_decomposition',
                                        'mse_amp mse_pha rsq cr')

    return mse_decomp(mse_amp, mse_pha, rsq, cr)


def invert_warping(fdatagrid, *, eval_points=None):
    r"""Compute the inverse of a diffeomorphism.

    Let :math:`\gamma : [a,b] \\rightarrow [a,b]` be a function strictly
    increasing, calculates the corresponding inverse
    :math:`\gamma^{-1} : [a,b] \\rightarrow [a,b]` such that
    :math:`\gamma^{-1} \circ \gamma = \gamma \circ \gamma^{-1} = \gamma_{id}`.

    Uses a PCHIP interpolator to compute approximately the inverse.

    Args:
        fdatagrid (:class:`FDataGrid`): Functions to be inverted.
        eval_points: (array_like, optional): Set of points where the
            functions are interpolated to obtain the inverse, by default uses
            the sample points of the fdatagrid.

    Returns:
        :class:`FDataGrid`: Inverse of the original functions.

    Raises:
        ValueError: If the functions are not strictly increasing or are
            multidimensional.

    Examples:

        >>> import numpy as np
        >>> from skfda import FDataGrid
        >>> from skfda.preprocessing.registration import invert_warping

        We will construct the warping :math:`\gamma : [0,1] \\rightarrow [0,1]`
        wich maps t to t^3.

        >>> t = np.linspace(0, 1)
        >>> gamma = FDataGrid(t**3, t)
        >>> gamma
        FDataGrid(...)

        We will compute the inverse.

        >>> inverse = invert_warping(gamma)
        >>> inverse
        FDataGrid(...)

        The result of the composition should be approximately the identity
        function .

        >>> identity = gamma.compose(inverse)
        >>> identity([0, 0.25, 0.5, 0.75, 1]).round(3)
        array([[ 0.  ,  0.25,  0.5 ,  0.75,  1.  ]])

    """

    if fdatagrid.ndim_image != 1 or fdatagrid.ndim_domain != 1:
        raise ValueError("Multidimensional object not supported.")

    if eval_points is None:
        eval_points = fdatagrid.sample_points[0]

    y = fdatagrid(eval_points, keepdims=False)

    data_matrix = numpy.empty((fdatagrid.nsamples, len(eval_points)))

    for i in range(fdatagrid.nsamples):
        data_matrix[i] = PchipInterpolator(y[i], eval_points)(eval_points)

    return fdatagrid.copy(data_matrix=data_matrix, sample_points=eval_points)


def _normalize_scale(t, a=0, b=1):
    """Perfoms an afine translation to normalize an interval.

    Args:
        t (numpy.ndarray): Array of dim 1 or 2 with at least 2 values.
        a (float): Starting point of the new interval. Defaults 0.
        b (float): Stopping point of the new interval. Defaults 1.

    Returns:
        (numpy.ndarray): Array with the transformed interval.
    """

    t = t.T  # Broadcast to normalize multiple arrays
    t1 = (t - t[0]).astype(float)  # Translation to [0, t[-1] - t[0]]
    t1 *= (b - a) / (t[-1] - t[0])  # Scale to [0, b-a]
    t1 += a  # Translation to [a, b]
    t1[0] = a  # Fix possible round errors
    t1[-1] = b

    return t1.T


def normalize_warping(warping, domain_range=None):
    """Rescale a warping to normalize their domain.

    Given a set of warpings :math:`\\gamma_i:[a,b] \\rightarrow [a,b]` it is
    used an affine traslation to change the domain of the transformation to
    other domain, :math:`\\hat \\gamma_i:[\\hat a,\\hat b] \\rightarrow
    [\\hat a, \\hat b]`.

    Args:
        warping (:class:`FDatagrid`): Set of warpings to rescale.
        domain_range (tuple, optional): New domain range of the warping. By
            default it is used the same domain range.
    Return:
        (:class:`FDataGrid`): FDataGrid with the warpings normalized.

    """

    if domain_range is None:
        domain_range = warping.domain_range[0]

    data_matrix = _normalize_scale(warping.data_matrix[..., 0], *domain_range)
    sample_points = _normalize_scale(warping.sample_points[0], *domain_range)

    return warping.copy(data_matrix=data_matrix, sample_points=sample_points,
                        domain_range=domain_range)



"""Functional Data Boxplot Module.

This module contains the classes to construct the functional data boxplot and
visualize it.

"""
import matplotlib
import matplotlib.pyplot as plt
import math

import numpy as np

from skfda.exploratory.depth import modified_band_depth
from ... import FDataGrid
from io import BytesIO
from abc import ABC, abstractmethod

__author__ = "Amanda Hernando Bernabé"
__email__ = "amanda.hernando@estudiante.uam.es"

class FDataBoxplot(ABC):
    """Abstract class inherited by the Boxplot and SurfaceBoxplot classes.

    It the data of the functional boxplot or surface boxplot of a FDataGrid object,
    depending on the dimensions of the domain, 1 or 2 respectively.

    It forces to both classes, Boxplot and SurfaceBoxplot to conain at least the median,
    central and outlying envelopes and a colormap for their graphical representation,
    obtained calling the plot method.

    """
    @abstractmethod
    def __init__(self, factor=1.5):
        if factor < 0:
            raise ValueError(
                "The number used to calculate the outlying envelope must be positive.")
        self._factor = factor

    @property
    def factor(self):
        return self._factor

    @property
    def fdatagrid(self):
        pass

    @property
    def median(self):
        pass

    @property
    def central_envelope(self):
        pass

    @property
    def outlying_envelope(self):
        pass

    @property
    def colormap(self):
        return self._colormap

    @colormap.setter
    def colormap(self, value):
        if not isinstance(value, matplotlib.colors.LinearSegmentedColormap):
            raise ValueError(
                "colormap must be of type matplotlib.colors.LinearSegmentedColormap")
        self._colormap = value

    @abstractmethod
    def plot(self, fig=None, ax=None, nrows=None, ncols=None):
        pass

    def _repr_svg_(self):
        plt.figure()
        fig, _ = self.plot()
        output = BytesIO()
        fig.savefig(output, format='svg')
        data = output.getvalue()
        plt.close(fig)
        return data.decode('utf-8')


class Boxplot(FDataBoxplot):
    r"""Representation of the functional boxplot.

    Class implementing the functionl boxplot which is an informative exploratory
    tool for visualizing functional data, as well as its generalization, the
    enhanced functional boxplot. Only supports 1 dimensional domain functional data.

    Based on the center outward ordering induced by a :ref:`depth measure <depth-measures>`
    for functional data, the descriptive statistics of a functional boxplot are: the
    envelope of the 50% central region, the median curve,and the maximum non-outlying envelope.
    In addition, outliers can be detected in a functional boxplot by the 1.5 times the 50%
    central region empirical rule, analogous to the rule for classical boxplots.

    Attributes:
        fdatagrid (FDataGrid): Object containing the data.
        median (array, (fdatagrid.ndim_image, nsample_points)): contains
            the median/s.
        central_envelope (array, (fdatagrid.ndim_image, 2, nsample_points)):
            contains the central envelope/s.
        outlying_envelope (array, (fdatagrid.ndim_image, 2, nsample_points)):
            contains the outlying envelope/s.
        colormap (matplotlib.colors.LinearSegmentedColormap): Colormap from
            which the colors to represent the central regions are selected.
        central_regions (array, (fdatagrid.ndim_image * ncentral_regions, 2,
            nsample_points)): contains the central regions.
        outliers (array, (fdatagrid.ndim_image, fdatagrid.nsamples)):
            contains the outliers
        barcol (string): Color of the envelopes and vertical lines.
        outliercol (string): Color of the ouliers.
        mediancol (string): Color of the median.
        show_full_outliers (boolean): If False (the default) then only the part
            outside the box is plotted. If True, complete outling curves are plotted

    Example:
        Function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2], [0.5, 0.5, 1, 2, 1.5, 1], [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> sample_points = [0, 2, 4, 6, 8, 10]
        >>> fd = FDataGrid(data_matrix, sample_points, dataset_label="dataset", axes_labels=["x_label", "y_label"])
        >>> Boxplot(fd)
        Boxplot(
            FDataGrid=FDataGrid(
                array([[[ 1. ],
                        [ 1. ],
                        [ 2. ],
                        [ 3. ],
                        [ 2.5],
                        [ 2. ]],
        <BLANKLINE>
                       [[ 0.5],
                        [ 0.5],
                        [ 1. ],
                        [ 2. ],
                        [ 1.5],
                        [ 1. ]],
        <BLANKLINE>
                       [[-1. ],
                        [-1. ],
                        [-0.5],
                        [ 1. ],
                        [ 1. ],
                        [ 0.5]],
        <BLANKLINE>
                       [[-0.5],
                        [-0.5],
                        [-0.5],
                        [-1. ],
                        [-1. ],
                        [-1. ]]]),
                sample_points=[array([ 0,  2,  4,  6,  8, 10])],
                domain_range=array([[ 0, 10]]),
                dataset_label='dataset',
                axes_labels=['x_label', 'y_label'],
                extrapolation=None,
                interpolator=SplineInterpolator(interpolation_order=1, smoothness_parameter=0.0, monotone=False),
                keepdims=False),
            median=array([[ 0.5,  0.5,  1. ,  2. ,  1.5,  1. ]]),
            central envelope=array([[[ 0.5,  0.5,  1. ,  2. ,  1.5,  1. ],
                    [-1. , -1. , -0.5,  1. ,  1. ,  0.5]]]),
            outlying envelope=array([[[ 1.  ,  1.  ,  2.  ,  3.  ,  2.25,  1.75],
                    [-1.  , -1.  , -0.5 , -0.5 ,  0.25, -0.25]]]),
            central_regions=array([[[ 0.5,  0.5,  1. ,  2. ,  1.5,  1. ],
                    [-1. , -1. , -0.5,  1. ,  1. ,  0.5]]]),
            outliers=array([[ 1.,  0.,  0.,  1.]]))

    """

    def __init__(self, fdatagrid, method=modified_band_depth, prob=[0.5],
                 factor=1.5):
        """Initialization of the Boxplot class.

        Args:
            fdatagrid (FDataGrid): Object containing the data.
            method (:ref:`depth measure <depth-measures>`, optional): Method
                used to order the data. Defaults to :func:`modified band depth
                <fda.depth_measures.modified_band_depth>`.
            prob (list of float, optional): List with float numbers (in the range
                from 1 to 0) that indicate which central regions to represent.
                Defaults to [0.5] which represents the 50% central region.
            factor (double): Number used to calculate the outlying envelope.

        """
        FDataBoxplot.__init__(self, factor)

        if fdatagrid.ndim_domain != 1:
            raise ValueError(
                "Function only supports FDataGrid with domain dimension 1.")

        if sorted(prob, reverse=True) != prob:
            raise ValueError(
                "Probabilities required to be in descending order.")

        if min(prob) < 0 or max(prob) > 1:
            raise ValueError("Probabilities must be between 0 and 1.")

        nsample_points = len(fdatagrid.sample_points[0])
        ncentral_regions = len(prob)

        self._median = np.ndarray((fdatagrid.ndim_image, nsample_points))
        self._central_envelope = np.ndarray((fdatagrid.ndim_image, 2,
                                             nsample_points))
        self._outlying_envelope = np.ndarray((fdatagrid.ndim_image, 2,
                                              nsample_points))
        self._central_regions = np.ndarray(
            (fdatagrid.ndim_image * ncentral_regions,
             2, nsample_points))
        self._outliers = np.zeros((fdatagrid.ndim_image, fdatagrid.nsamples))

        depth = method(fdatagrid)
        indices_descencing_depth = (-depth).argsort(axis=0)

        for m in range(fdatagrid.ndim_image):

            for i in range(len(prob)):

                indices_samples = indices_descencing_depth[:, m][
                                  :math.ceil(fdatagrid.nsamples * prob[i])]
                samples_used = fdatagrid.data_matrix[indices_samples, :, m]
                max_samples_used = np.amax(samples_used, axis=0)
                min_samples_used = np.amin(samples_used, axis=0)

                if prob[i] == 0.5:
                    # central envelope
                    self._central_envelope[m] = np.asarray(
                        [max_samples_used.T, min_samples_used.T])

                    # outlying envelope
                    max_value = np.amax(fdatagrid.data_matrix[:, :, m], axis=0)
                    min_value = np.amin(fdatagrid.data_matrix[:, :, m], axis=0)
                    iqr = np.absolute(max_samples_used - min_samples_used)
                    outlying_max_envelope = np.minimum(
                        max_samples_used + iqr * factor, max_value)
                    outlying_min_envelope = np.maximum(
                        min_samples_used - iqr * factor, min_value)
                    self._outlying_envelope[m] = np.asarray(
                        [outlying_max_envelope.flatten(),
                         outlying_min_envelope.flatten()])

                    # outliers
                    for j in list(range(fdatagrid.nsamples)):
                        outliers_above = (
                                outlying_max_envelope < fdatagrid.data_matrix[
                                                        j, :, m])
                        outliers_below = (
                                outlying_min_envelope > fdatagrid.data_matrix[
                                                        j, :, m])
                        if (
                                outliers_above.sum() > 0 or outliers_below.sum() > 0):
                            self._outliers[m, j] = 1
                # central regions
                self._central_regions[ncentral_regions * m + i] = np.asarray(
                    [max_samples_used.flatten(), min_samples_used.flatten()])

            # mean sample
            self._median[m] = fdatagrid.data_matrix[
                              indices_descencing_depth[0, m], :, m].T

        self._fdatagrid = fdatagrid
        self._prob = prob
        self._colormap = plt.cm.get_cmap('RdPu')
        self.barcol = "blue"
        self.outliercol = "red"
        self.mediancol = "black"
        self._show_full_outliers = False

    @property
    def fdatagrid(self):
        return self._fdatagrid

    @property
    def median(self):
        return self._median

    @property
    def central_envelope(self):
        return self._central_envelope

    @property
    def outlying_envelope(self):
        return self._outlying_envelope

    @property
    def central_regions(self):
        return self._central_regions

    @property
    def outliers(self):
        return self._outliers

    @property
    def show_full_outliers(self):
        return self._show_full_outliers

    @show_full_outliers.setter
    def show_full_outliers(self, boolean):
        if not isinstance(boolean, bool):
            raise ValueError("show_full_outliers must be boolean type")
        self._show_full_outliers = boolean

    def plot(self, fig=None, ax=None, nrows=None, ncols=None):
        """Visualization of the functional boxplot of the fdatagrid (ndim_domain=1).

        Args:
            fig (figure object, optional): figure over with the graphs are
                plotted in case ax is not specified. If None and ax is also None,
                the figure is initialized.
            ax (list of axis objects, optional): axis over where the graphs are
                plotted. If None, see param fig.
            nrows(int, optional): designates the number of rows of the figure
                to plot the different dimensions of the image. Only specified
                if fig and ax are None.
            ncols(int, optional): designates the number of columns of the figure
                to plot the different dimensions of the image. Only specified
                if fig and ax are None.

         Returns:
            fig (figure object): figure object in which the graphs are plotted.
            ax (axes object): axes in which the graphs are plotted.

        """

        fig, ax = self.fdatagrid.generic_plotting_checks(fig, ax, nrows,
                                                         ncols)
        tones = np.linspace(0.1, 1.0, len(self._prob) + 1, endpoint=False)[1:]
        color = self.colormap(tones)

        if self.show_full_outliers:
            var_zorder = 1
        else:
            var_zorder = 4

        for m in range(self.fdatagrid.ndim_image):

            # outliers
            for j in list(range(self.fdatagrid.nsamples)):
                if self.outliers[m, j]:
                    ax[m].plot(self.fdatagrid.sample_points[0],
                               self.fdatagrid.data_matrix[j, :, m],
                               color=self.outliercol,
                               linestyle='--', zorder=1)

            for i in range(len(self._prob)):
                # central regions
                ax[m].fill_between(self.fdatagrid.sample_points[0],
                                   self.central_regions[
                                       m * len(self._prob) + i, 0],
                                   self.central_regions[
                                       m * len(self._prob) + i, 1],
                                   facecolor=color[i], zorder=var_zorder)

            # outlying envelope
            ax[m].plot(self.fdatagrid.sample_points[0],
                       self.outlying_envelope[m, 0],
                       self.fdatagrid.sample_points[0],
                       self.outlying_envelope[m, 1], color=self.barcol,
                       zorder=4)

            # central envelope
            ax[m].plot(self.fdatagrid.sample_points[0],
                       self.central_envelope[m, 0],
                       self.fdatagrid.sample_points[0],
                       self.central_envelope[m, 1], color=self.barcol,
                       zorder=4)

            # vertical lines
            index = math.ceil(self.fdatagrid.ncol / 2)
            x = self.fdatagrid.sample_points[0][index]
            ax[m].plot([x, x], [self.outlying_envelope[m, 0][index],
                                self.central_envelope[m, 0][index]],
                       color=self.barcol,
                       zorder=4)
            ax[m].plot([x, x], [self.outlying_envelope[m, 1][index],
                                self.central_envelope[m, 1][index]],
                       color=self.barcol, zorder=4)

            # median sample
            ax[m].plot(self.fdatagrid.sample_points[0], self.median[m],
                       color=self.mediancol, zorder=5)

        self.fdatagrid.set_labels(fig, ax)

        return fig, ax

    def __repr__(self):
        """Return repr(self)."""
        return (f"Boxplot("
                f"\nFDataGrid={repr(self.fdatagrid)},"
                f"\nmedian={repr(self.median)},"
                f"\ncentral envelope={repr(self.central_envelope)},"
                f"\noutlying envelope={repr(self.outlying_envelope)},"
                f"\ncentral_regions={repr(self.central_regions)},"
                f"\noutliers={repr(self.outliers)})").replace('\n', '\n    ')


class SurfaceBoxplot(FDataBoxplot):
    r"""Representation of the surface boxplot.

    Class implementing the surface boxplot. Analogously to the functional boxplot,
    it is an informative exploratory tool for visualizing functional data with
    domain dimension 2. Nevertheless, it does not implement the enhanced
    surface boxplot.

    Based on the center outward ordering induced by a :ref:`depth measure <depth-measures>`
    for functional data, it represents the envelope of the 50% central region, the median curve,
    and the maximum non-outlying envelope.

    Attributes:
        fdatagrid (FDataGrid): Object containing the data.
        median (array, (fdatagrid.ndim_image, lx, ly)): contains
            the median/s.
        central_envelope (array, (fdatagrid.ndim_image, 2, lx, ly)):
            contains the central envelope/s.
        outlying_envelope (array,(fdatagrid.ndim_image, 2, lx, ly)):
            contains the outlying envelope/s.
        colormap (matplotlib.colors.LinearSegmentedColormap): Colormap from
            which the colors to represent the central regions are selected.
        boxcol (string): Color of the box, which includes median and central envelope.
        outcol (string): Color of the outlying envelope.

    Example:
        Function :math:`f : \mathbb{R^2}\longmapsto\mathbb{R^2}`.

        >>> data_matrix = [[[[1, 4], [0.3, 1.5], [1, 3]], [[2, 8], [0.4, 2], [2, 9]]],
        ...                [[[2, 10], [0.5, 3], [2, 10]], [[3, 12], [0.6, 3], [3, 15]]]]
        >>> sample_points = [[2, 4], [3, 6, 8]]
        >>> fd = FDataGrid(data_matrix, sample_points, dataset_label= "dataset",
        ...                axes_labels=["x1_label", "x2_label", "y1_label", "y2_label"])
        >>> SurfaceBoxplot(fd)
        SurfaceBoxplot(
            FDataGrid=FDataGrid(
                array([[[[  1. ,   4. ],
                         [  0.3,   1.5],
                         [  1. ,   3. ]],
        <BLANKLINE>
                        [[  2. ,   8. ],
                         [  0.4,   2. ],
                         [  2. ,   9. ]]],
        <BLANKLINE>
        <BLANKLINE>
                       [[[  2. ,  10. ],
                         [  0.5,   3. ],
                         [  2. ,  10. ]],
        <BLANKLINE>
                        [[  3. ,  12. ],
                         [  0.6,   3. ],
                         [  3. ,  15. ]]]]),
                sample_points=[array([2, 4]), array([3, 6, 8])],
                domain_range=array([[2, 4],
                       [3, 8]]),
                dataset_label='dataset',
                axes_labels=['x1_label', 'x2_label', 'y1_label', 'y2_label'],
                extrapolation=None,
                interpolator=SplineInterpolator(interpolation_order=1, smoothness_parameter=0.0, monotone=False),
                keepdims=False),
            median=array([[[ 1. ,  0.3,  1. ],
                    [ 2. ,  0.4,  2. ]],
        <BLANKLINE>
                   [[ 4. ,  1.5,  3. ],
                    [ 8. ,  2. ,  9. ]]]),
            central envelope=array([[[[ 1. ,  0.3,  1. ],
                     [ 2. ,  0.4,  2. ]],
        <BLANKLINE>
                    [[ 1. ,  0.3,  1. ],
                     [ 2. ,  0.4,  2. ]]],
        <BLANKLINE>
        <BLANKLINE>
                   [[[ 4. ,  1.5,  3. ],
                     [ 8. ,  2. ,  9. ]],
        <BLANKLINE>
                    [[ 4. ,  1.5,  3. ],
                     [ 8. ,  2. ,  9. ]]]]),
            outlying envelope=array([[[[ 1. ,  0.3,  1. ],
                     [ 2. ,  0.4,  2. ]],
        <BLANKLINE>
                    [[ 1. ,  0.3,  1. ],
                     [ 2. ,  0.4,  2. ]]],
        <BLANKLINE>
        <BLANKLINE>
                   [[[ 4. ,  1.5,  3. ],
                     [ 8. ,  2. ,  9. ]],
        <BLANKLINE>
                    [[ 4. ,  1.5,  3. ],
                     [ 8. ,  2. ,  9. ]]]]))


    """

    def __init__(self, fdatagrid, method=modified_band_depth, factor=1.5):
        """Initialization of the functional boxplot.

        Args:
            fdatagrid (FDataGrid): Object containing the data.
            method (:ref:`depth measure <depth-measures>`, optional): Method
                used to order the data. Defaults to :func:`modified band depth
                <fda.depth_measures.modified_band_depth>`.
            prob (list of float, optional): List with float numbers (in the range
                from 1 to 0) that indicate which central regions to represent.
                Defaults to [0.5] which represents the 50% central region.
            factor (double): Number used to calculate the outlying envelope.

        """
        FDataBoxplot.__init__(self, factor)

        if fdatagrid.ndim_domain != 2:
            raise ValueError(
                "Class only supports FDataGrid with domain dimension 2.")

        lx = len(fdatagrid.sample_points[0])
        ly = len(fdatagrid.sample_points[1])

        self._median = np.ndarray((fdatagrid.ndim_image, lx, ly))
        self._central_envelope = np.ndarray((fdatagrid.ndim_image, 2, lx, ly))
        self._outlying_envelope = np.ndarray((fdatagrid.ndim_image, 2, lx, ly))

        depth = method(fdatagrid)
        indices_descencing_depth = (-depth).argsort(axis=0)

        for m in range(fdatagrid.ndim_image):
            indices_samples = indices_descencing_depth[:, m][
                              :math.ceil(fdatagrid.nsamples * 0.5)]
            samples_used = fdatagrid.data_matrix[indices_samples, :, :, m]
            max_samples_used = np.amax(samples_used, axis=0)
            min_samples_used = np.amin(samples_used, axis=0)

            # mean sample
            self._median[m] = fdatagrid.data_matrix[
                              indices_descencing_depth[0, m], :, :, m]

            # central envelope
            self._central_envelope[m] = np.asarray([max_samples_used,
                                                    min_samples_used])

            # outlying envelope
            max_value = np.amax(fdatagrid.data_matrix[:, :, :, m], axis=0)
            min_value = np.amin(fdatagrid.data_matrix[:, :, :, m], axis=0)
            iqr = np.absolute(max_samples_used - min_samples_used)
            oulying_max_envelope = np.minimum(max_samples_used + iqr * factor,
                                              max_value)
            oulying_min_envelope = np.maximum(min_samples_used - iqr * factor,
                                              min_value)
            self._outlying_envelope[m] = np.asarray([oulying_max_envelope,
                                                     oulying_min_envelope])

        self._fdatagrid = fdatagrid
        self.colormap = plt.cm.get_cmap('Greys')
        self._boxcol = 1.0
        self._outcol = 0.7

    @property
    def fdatagrid(self):
        return self._fdatagrid

    @property
    def median(self):
        return self._median

    @property
    def central_envelope(self):
        return self._central_envelope

    @property
    def outlying_envelope(self):
        return self._outlying_envelope

    @property
    def boxcol(self):
        return self._boxcol

    @boxcol.setter
    def boxcol(self, value):
        if value < 0 or value > 1:
            raise ValueError(
                "boxcol must be a number between 0 and 1.")

        self._boxcol = value

    @property
    def outcol(self):
        return self._outcol

    @outcol.setter
    def outcol(self, value):
        if value < 0 or value > 1:
            raise ValueError(
                "outcol must be a number between 0 and 1.")
        self._outcol = value

    def plot(self, fig=None, ax=None, nrows=None, ncols=None):
        """Visualization of the surface boxplot of the fdatagrid (ndim_domain=2).

         Args:
             fig (figure object, optional): figure over with the graphs are
                plotted in case ax is not specified. If None and ax is also None,
                the figure is initialized.
             ax (list of axis objects, optional): axis over where the graphs are
                plotted. If None, see param fig.
             nrows(int, optional): designates the number of rows of the figure
                 to plot the different dimensions of the image. Only specified
                 if fig and ax are None.
             ncols(int, optional): designates the number of columns of the figure
                 to plot the different dimensions of the image. Only specified
                 if fig and ax are None.

        Returns:
             fig (figure object): figure object in which the graphs are plotted.
             ax (axes object): axes in which the graphs are plotted.

        """
        fig, ax = self.fdatagrid.generic_plotting_checks(fig, ax, nrows,
                                                         ncols)
        x = self.fdatagrid.sample_points[0]
        lx = len(x)
        y = self.fdatagrid.sample_points[1]
        ly = len(y)
        X, Y = np.meshgrid(x, y)

        for m in range(self.fdatagrid.ndim_image):

            # mean sample
            ax[m].plot_wireframe(X, Y, np.squeeze(self.median[m]).T,
                                 rstride=ly, cstride=lx,
                                 color=self.colormap(self.boxcol))
            ax[m].plot_surface(X, Y, np.squeeze(self.median[m]).T,
                               color=self.colormap(self.boxcol), alpha=0.8)

            # central envelope
            ax[m].plot_surface(X, Y, np.squeeze(self.central_envelope[m, 0]).T,
                               color=self.colormap(self.boxcol), alpha=0.5)
            ax[m].plot_wireframe(X, Y,
                                 np.squeeze(self.central_envelope[m, 0]).T,
                                 rstride=ly, cstride=lx,
                                 color=self.colormap(self.boxcol))
            ax[m].plot_surface(X, Y, np.squeeze(self.central_envelope[m, 1]).T,
                               color=self.colormap(self.boxcol), alpha=0.5)
            ax[m].plot_wireframe(X, Y,
                                 np.squeeze(self.central_envelope[m, 1]).T,
                                 rstride=ly, cstride=lx,
                                 color=self.colormap(self.boxcol))

            # box vertical lines
            for indices in [(0, 0), (0, ly - 1), (lx - 1, 0),
                            (lx - 1, ly - 1)]:
                x_corner = x[indices[0]]
                y_corner = y[indices[1]]
                ax[m].plot([x_corner, x_corner], [y_corner, y_corner],
                           [self.central_envelope[
                                m, 1, indices[0], indices[1]],
                            self.central_envelope[
                                m, 0, indices[0], indices[1]]],
                           color=self.colormap(self.boxcol))

            # outlying envelope
            ax[m].plot_surface(X, Y,
                               np.squeeze(self.outlying_envelope[m, 0]).T,
                               color=self.colormap(self.outcol), alpha=0.3)
            ax[m].plot_wireframe(X, Y,
                                 np.squeeze(self.outlying_envelope[m, 0]).T,
                                 rstride=ly, cstride=lx,
                                 color=self.colormap(self.outcol))
            ax[m].plot_surface(X, Y,
                               np.squeeze(self.outlying_envelope[m, 1]).T,
                               color=self.colormap(self.outcol), alpha=0.3)
            ax[m].plot_wireframe(X, Y,
                                 np.squeeze(self.outlying_envelope[m, 1]).T,
                                 rstride=ly, cstride=lx,
                                 color=self.colormap(self.outcol))

            # vertical lines from central to outlying envelope
            x_index = math.floor(lx / 2)
            x_central = x[x_index]
            y_index = math.floor(ly / 2)
            y_central = y[y_index]
            ax[m].plot([x_central, x_central], [y_central, y_central],
                       [self.outlying_envelope[m, 1, x_index, y_index],
                        self.central_envelope[m, 1, x_index, y_index]],
                       color=self.colormap(self.boxcol))
            ax[m].plot([x_central, x_central], [y_central, y_central],
                       [self.outlying_envelope[m, 0, x_index, y_index],
                        self.central_envelope[m, 0, x_index, y_index]],
                       color=self.colormap(self.boxcol))

        self.fdatagrid.set_labels(fig, ax)

        return fig, ax

    def __repr__(self):
        """Return repr(self)."""
        return (f"SurfaceBoxplot("
                f"\nFDataGrid={repr(self.fdatagrid)},"
                f"\nmedian={repr(self.median)},"
                f"\ncentral envelope={repr(self.central_envelope)},"
                f"\noutlying envelope={repr(self.outlying_envelope)})").replace('\n', '\n    ')



"""
Module to interpolate functional data objects.
"""

from abc import ABC, abstractmethod

import numpy as np

# Scipy interpolator methods used internally
from scipy.interpolate import (PchipInterpolator, UnivariateSpline,
                               RectBivariateSpline, RegularGridInterpolator)

from .._utils import _list_of_arrays
from .evaluator import Evaluator, EvaluatorConstructor


class SplineInterpolator(EvaluatorConstructor):
    r"""Spline interpolator of :class:`FDataGrid`.

    Spline interpolator of discretized functional objects. Implements different
    interpolation methods based in splines, using the sample points of the
    grid as nodes to interpolate.

    See the interpolation example to a detailled explanation.

    Attributes:
        interpolator_order (int, optional): Order of the interpolation, 1
            for linear interpolation, 2 for cuadratic, 3 for cubic and so
            on. In case of curves and surfaces there is available
            interpolation up to degree 5. For higher dimensional objects
            only linear or nearest interpolation is available. Default
            lineal interpolation.
        smoothness_parameter (float, optional): Penalisation to perform
            smoothness interpolation. Option only available for curves and
            surfaces. If 0 the residuals of the interpolation will be 0.
            Defaults 0.
        monotone (boolean, optional): Performs monotone interpolation in
            curves using a PCHIP interpolator. Only valid for curves (domain
            dimension equal to 1) and interpolation order equal to 1 or 3.
            Defaults false.

    """

    def __init__(self, interpolation_order=1, smoothness_parameter=0.,
                 monotone=False):
        r"""Constructor of the SplineInterpolator.

        Args:
            interpolator_order (int, optional): Order of the interpolation, 1
                for linear interpolation, 2 for cuadratic, 3 for cubic and so
                on. In case of curves and surfaces there is available
                interpolation up to degree 5. For higher dimensional objects
                only linear or nearest interpolation is available. Default
                lineal interpolation.
            smoothness_parameter (float, optional): Penalisation to perform
                smoothness interpolation. Option only available for curves and
                surfaces. If 0 the residuals of the interpolation will be 0.
                Defaults 0.
            monotone (boolean, optional): Performs monotone interpolation in
                curves using a PCHIP interpolator. Only valid for curves (domain
                dimension equal to 1) and interpolation order equal to 1 or 3.
                Defaults false.

        """
        self._interpolation_order = interpolation_order
        self._smoothness_parameter = smoothness_parameter
        self._monotone = monotone

    @property
    def interpolation_order(self):
        "Returns the interpolation order"
        return self._interpolation_order

    @property
    def smoothness_parameter(self):
        "Returns the smoothness parameter"
        return self._smoothness_parameter

    @property
    def monotone(self):
        "Returns flag to perform monotone interpolation"
        return self._monotone

    def __eq__(self, other):
        """Equality operator between SplineInterpolator"""
        return (super().__eq__(other) and
                self.interpolation_order == other.interpolation_order and
                self.smoothness_parameter == other.smoothness_parameter and
                self.monoton == other.monotone)

    def evaluator(self, fdatagrid):
        """Construct a SplineInterpolatorEvaluator used in the evaluation.

        Args:
            fdatagrid (:class:`FDataGrid`): Functional object where the
                evaluator will be used.

        Returns:
            (:class:`SplineInterpolatorEvaluator`): Evaluator of the fdatagrid.

        """
        return SplineInterpolatorEvaluator(fdatagrid, self.interpolation_order,
                                           self.smoothness_parameter,
                                           self.monotone)

    def __repr__(self):
        """repr method of the interpolator"""
        return (f"{type(self).__name__}("
                f"interpolation_order={self.interpolation_order}, "
                f"smoothness_parameter={self.smoothness_parameter}, "
                f"monotone={self.monotone})")


class SplineInterpolatorEvaluator(Evaluator):
    r"""Spline interpolator evaluator of :class:`FDataGrid`.

    It is generated by the SplineInterpolator, and it is used internally
    during the evaluation.

    Spline interpolator of discretized functional objects. Implements different
    interpolation methods based in splines, using the sample points of the
    grid as nodes to interpolate.

    See the interpolation example to a detailled explanation.

    Attributes:
        interpolator_order (int, optional): Order of the interpolation, 1
            for linear interpolation, 2 for cuadratic, 3 for cubic and so
            on. In case of curves and surfaces there is available
            interpolation up to degree 5. For higher dimensional objects
            only linear or nearest interpolation is available. Default
            lineal interpolation.
        smoothness_parameter (float, optional): Penalisation to perform
            smoothness interpolation. Option only available for curves and
            surfaces. If 0 the residuals of the interpolation will be 0.
            Defaults 0.
        monotone (boolean, optional): Performs monotone interpolation in
            curves using a PCHIP interpolator. Only valid for curves (domain
            dimension equal to 1) and interpolation order equal to 1 or 3.
            Defaults false.

    """

    def __init__(self, fdatagrid, k=1, s=0., monotone=False):
        r"""Constructor of the SplineInterpolatorEvaluator.

        Args:
            fdatagir (fdatagrid): Grid to be interpolated.
            interpolator_order (int, optional): Order of the interpolation, 1
                for linear interpolation, 2 for cuadratic, 3 for cubic and so
                on. In case of curves and surfaces there is available
                interpolation up to degree 5. For higher dimensional objects
                only linear or nearest interpolation is available. Default
                lineal interpolation.
            smoothness_parameter (float, optional): Penalisation to perform
                smoothness interpolation. Option only available for curves and
                surfaces. If 0 the residuals of the interpolation will be 0.
                Defaults 0.
            monotone (boolean, optional): Performs monotone interpolation in
                curves using a PCHIP interpolator. Only valid for curves (domain
                dimension equal to 1) and interpolation order equal to 1 or 3.
                Defaults false.

        """
        sample_points = fdatagrid.sample_points
        data_matrix = fdatagrid.data_matrix

        self._fdatagrid = fdatagrid
        self._ndim_image = fdatagrid.ndim_image
        self._ndim_domain = fdatagrid.ndim_domain
        self._nsamples = fdatagrid.nsamples
        self._keepdims = fdatagrid.keepdims
        self._domain_range = fdatagrid.domain_range

        if self._ndim_domain == 1:
            self._splines = self._construct_spline_1_m(sample_points,
                                                       data_matrix,
                                                       k, s, monotone)
        elif monotone:
            raise ValueError("Monotone interpolation is only supported with "
                             "domain dimension equal to 1.")

        elif self._ndim_domain == 2:
            self._splines = self._construct_spline_2_m(sample_points,
                                                       data_matrix, k, s)

        elif s != 0:
            raise ValueError("Smoothing interpolation is only supported with "
                             "domain dimension up to 2, s should be 0.")

        else:
            self._splines = self._construct_spline_n_m(sample_points,
                                                       data_matrix, k)

        # After the creation of the splines the fdatagrid reference can
        # be deleted
        self._fdatagrid = None

    def _construct_spline_1_m(self, sample_points, data_matrix, k, s, monotone):
        r"""Construct the matrix of interpolators for curves.

        Constructs the matrix of interpolators for objects with domain
        dimension = 1. Calling internally during the creationg of the evaluator.

        Uses internally the scipy interpolator UnivariateSpline or
        PchipInterpolator.

        Args:
            sample_points (np.ndarray): Sample points of the fdatagrid.
            data_matrix (np.ndarray): Data matrix of the fdatagrid.
            k (integer): Order of the spline interpolators.

        Returns:
            (np.ndarray): Array of size nsamples x ndim_image with the
            corresponding interpolator of the sample i, and image dimension j
            in the entry (i,j) of the array.

        Raises:
            ValueError: If the value of the interpolation k is not valid.

        """
        if k > 5 or k < 1:
            raise ValueError(f"Invalid degree of interpolation ({k}). Must be "
                             f"an integer greater than 0 and lower or "
                             f"equal than 5.")

        if monotone and s != 0:
            raise ValueError("Smoothing interpolation is not supported with "
                             "monotone interpolation")

        if monotone and (k == 2 or k == 4):
            raise ValueError(f"monotone interpolation of degree {k}"
                             f"not supported.")

        # Monotone interpolation of degree 1 is performed with linear spline
        if monotone and k == 1:
            monotone = False

        # Evaluator of splines called in evaluate

        def _spline_evaluator_1_m(spl, t, der):

            return spl(t, der)

        def _process_derivative_1_m(derivative):

            return derivative

        self._spline_evaluator = _spline_evaluator_1_m

        self._process_derivative = _process_derivative_1_m

        sample_points = sample_points[0]

        if monotone:
            def constructor(data):
                """Constructs an unidimensional cubic monotone interpolator"""
                return PchipInterpolator(sample_points, data)

        else:

            def constructor(data):
                """Constructs an unidimensional interpolator"""
                return UnivariateSpline(sample_points, data, s=s, k=k)

        return np.apply_along_axis(constructor, 1, data_matrix)

    def _construct_spline_2_m(self, sample_points, data_matrix, k, s):
        r"""Construct the matrix of interpolators for surfaces.

        Constructs the matrix of interpolators for objects with domain
        dimension = 2. Calling internally during the creationg of the evaluator.

        Uses internally the scipy interpolator RectBivariateSpline.

        Args:
            sample_points (np.ndarray): Sample points of the fdatagrid.
            data_matrix (np.ndarray): Data matrix of the fdatagrid.
            k (integer): Order of the spline interpolators.

        Returns:
            (np.ndarray): Array of size nsamples x ndim_image with the
            corresponding interpolator of the sample i, and image dimension j
            in the entry (i,j) of the array.

        Raises:
            ValueError: If the value of the interpolation k is not valid.

        """
        if np.isscalar(k):
            kx = ky = k
        elif len(k) != 2:
            raise ValueError("k should be numeric or a tuple of length 2.")
        else:
            kx = k[0]
            ky = k[1]

        if kx > 5 or kx <= 0 or ky > 5 or ky <= 0:
            raise ValueError(f"Invalid degree of interpolation ({kx},{ky}). "
                             f"Must be an integer greater than 0 and lower or "
                             f"equal than 5.")

        def _spline_evaluator_2_m(spl, t, der):

            return spl(t[:, 0], t[:, 1], dx=der[0], dy=der[1], grid=False)

        def _process_derivative_2_m(derivative):
            if np.isscalar(derivative):
                derivative = 2*[derivative]
            elif len(derivative) != 2:
                raise ValueError("derivative should be a numeric value "
                                 "or a tuple of length 2 with (dx,dy).")

            return derivative

        # Evaluator of splines called in evaluate
        self._spline_evaluator = _spline_evaluator_2_m
        self._process_derivative = _process_derivative_2_m

        # Matrix of splines
        spline = np.empty((self._nsamples, self._ndim_image), dtype=object)

        for i in range(self._nsamples):
            for j in range(self._ndim_image):
                spline[i, j] = RectBivariateSpline(sample_points[0],
                                                   sample_points[1],
                                                   data_matrix[i, :, :, j],
                                                   kx=kx, ky=ky, s=s)

        return spline

    def _construct_spline_n_m(self, sample_points, data_matrix, k):
        r"""Construct the matrix of interpolators.

        Constructs the matrix of interpolators for objects with domain
        dimension > 2. Calling internally during the creationg of the evaluator.

        Only linear and nearest interpolators are available for objects with
        domain dimension >= 3. Uses internally the scipy interpolator
        RegularGridInterpolator.

        Args:
            sample_points (np.ndarray): Sample points of the fdatagrid.
            data_matrix (np.ndarray): Data matrix of the fdatagrid.
            k (integer): Order of the spline interpolators.

        Returns:
            (np.ndarray): Array of size nsamples x ndim_image with the
            corresponding interpolator of the sample i, and image dimension j
            in the entry (i,j) of the array.

        Raises:
            ValueError: If the value of the interpolation k is not valid.

        """
        # Parses method of interpolation
        if k == 0:
            method = 'nearest'
        elif k == 1:
            method = 'linear'
        else:
            raise ValueError("interpolation order should be 0 (nearest) or 1 "
                             "(linear).")

        def _process_derivative_n_m(derivative):
            if derivative != 0:
                raise ValueError("derivates not suported for functional data "
                                 " with domain dimension greater than 2.")

            return derivative

        def _spline_evaluator_n_m(spl, t, derivative):

            return spl(t)

        # Method to process derivative argument
        self._process_derivative = _process_derivative_n_m

        # Evaluator of splines called in evaluate
        self._spline_evaluator = _spline_evaluator_n_m

        spline = np.empty((self._nsamples, self._ndim_image), dtype=object)

        for i in range(self._nsamples):
            for j in range(self._ndim_image):
                spline[i, j] = RegularGridInterpolator(
                    sample_points, data_matrix[i, ..., j], method, False)

        return spline

    def evaluate(self, eval_points, *, derivative=0):
        r"""Evaluation method.

        Evaluates the samples at different evaluation points. The evaluation
        call will receive a 3-d array with the evaluation points for each sample.

        This method is called internally by :meth:`evaluate` when the argument
        `aligned_evaluation` is False.

        Args:
            eval_points (np.ndarray): Numpy array with shape
                `(n_samples, number_eval_points, ndim_domain)` with the
                 evaluation points for each sample.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (np.darray): Numpy 3d array with shape `(n_samples,
                number_eval_points, ndim_image)` with the result of the
                evaluation. The entry (i,j,k) will contain the value k-th image
                dimension of the i-th sample, at the j-th evaluation point.

        Raises:
            ValueError: In case of an incorrect value of the derivative
                argument.

        """
        derivative = self._process_derivative(derivative)

        # Constructs the evaluator for t_eval
        if self._ndim_image == 1:
            def evaluator(spl):
                """Evaluator of object with image dimension equal to 1."""
                return self._spline_evaluator(spl[0], eval_points, derivative)
        else:
            def evaluator(spl_m):
                """Evaluator of multimensional object"""
                return np.dstack(
                    [self._spline_evaluator(spl, eval_points, derivative)
                     for spl in spl_m]).flatten()

        # Points evaluated inside the domain
        res = np.apply_along_axis(evaluator, 1, self._splines)
        res = res.reshape(self._nsamples, eval_points.shape[0],
                          self._ndim_image)

        return res

    def evaluate_composed(self, eval_points, *, derivative=0):
        """Evaluation method.

        Evaluates the samples at different evaluation points. The evaluation
        call will receive a 3-d array with the evaluation points for each sample.

        This method is called internally by :meth:`evaluate` when the argument
        `aligned_evaluation` is False.

        Args:
            eval_points (np.ndarray): Numpy array with shape
                `(n_samples, number_eval_points, ndim_domain)` with the
                 evaluation points for each sample.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (np.darray): Numpy 3d array with shape `(n_samples,
                number_eval_points, ndim_image)` with the result of the
                evaluation. The entry (i,j,k) will contain the value k-th image
                dimension of the i-th sample, at the j-th evaluation point.

        Raises:
            ValueError: In case of an incorrect value of the derivative
                argument.

        """
        shape = (self._nsamples, eval_points.shape[1], self._ndim_image)
        res = np.empty(shape)

        derivative = self._process_derivative(derivative)

        if self._ndim_image == 1:
            def evaluator(t, spl):
                """Evaluator of sample with image dimension equal to 1"""
                return self._spline_evaluator(spl[0], t, derivative)

            for i in range(self._nsamples):
                res[i] = evaluator(eval_points[i], self._splines[i]).reshape(
                    (eval_points.shape[1], self._ndim_image))

        else:
            def evaluator(t, spl_m):
                """Evaluator of multidimensional sample"""
                return np.array([self._spline_evaluator(spl, t, derivative)
                                 for spl in spl_m]).T

            for i in range(self._nsamples):
                res[i] = evaluator(eval_points[i], self._splines[i])

        return res



import sklearn.utils
import numpy as np
from scipy.stats import multivariate_normal
import scipy.integrate
from ..misc import covariances
from .. import FDataGrid
from ..representation.interpolation import SplineInterpolator
from ..preprocessing.registration import normalize_warping


def make_gaussian_process(n_samples: int=100, n_features: int=100, *,
                          start: float=0., stop: float=1.,
                          mean=0, cov=None, noise: float=0.,
                          random_state=None):
    """Generate Gaussian process trajectories.

        Args:
            n_samples: The total number of trajectories.
            n_features: The total number of trajectories.
            start: Starting point of the trajectories.
            stop: Ending point of the trajectories.
            mean: The mean function of the process. Can be a callable accepting
                  a vector with the locations, or a vector with length
                  ``n_features``.
            cov: The covariance function of the process. Can be a
                  callable accepting two vectors with the locations, or a
                  matrix with size ``n_features`` x ``n_features``.
            noise: Standard deviation of Gaussian noise added to the data.
            random_state: Random state.

        Returns:
            :class:`FDataGrid` object comprising all the trajectories.

    """

    random_state = sklearn.utils.check_random_state(random_state)

    x = np.linspace(start, stop, n_features)

    if cov is None:
        cov = covariances.Brownian()

    covariance = covariances._execute_covariance(cov, x, x)

    if noise:
        covariance += np.eye(n_features) * noise ** 2

    mu = np.zeros(n_features)
    if callable(mean):
        mean = mean(x)
    mu += mean

    y = random_state.multivariate_normal(mu, covariance, n_samples)

    return FDataGrid(sample_points=x, data_matrix=y)


def make_sinusoidal_process(n_samples: int=15, n_features: int=100, *,
                            start: float=0., stop: float=1., period: float=1.,
                            phase_mean: float=0., phase_std: float=.6,
                            amplitude_mean: float=1., amplitude_std: float=.05,
                            error_std: float=.2, random_state=None):

    r"""Generate sinusoidal proccess.

    Each sample :math:`x_i(t)` is generated as:

    .. math::

        x_i(t) = \alpha_i \sin(\omega t + \phi_i) + \epsilon_i(t)


    where :math:`\omega=\frac{2 \pi}{\text{period}}`. Amplitudes
    :math:`\alpha_i` and phases :math:`\phi_i` are normally distributed.
    :math:`\epsilon_i(t)` is a gaussian white noise process.

    Args:
        n_samples: Total number of samples.
        n_features: Points per sample.
        start: Starting point of the samples.
        stop: Ending point of the samples.
        period: Period of the sine function.
        phase_mean: Mean of the phase.
        phase_std: Standard deviation of the phase.
        amplitude_mean: Mean of the amplitude.
        amplitude_std: Standard deviation of the amplitude.
        error_std: Standard deviation of the gaussian Noise.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the samples.

    """

    random_state = sklearn.utils.check_random_state(random_state)

    t = np.linspace(start, stop, n_features)

    alpha = np.diag(random_state.normal(amplitude_mean, amplitude_std,
                                        n_samples))

    phi = np.outer(random_state.normal(phase_mean, phase_std,  n_samples),
                   np.ones(n_features))

    error = random_state.normal(0, error_std, (n_samples, n_features))


    y = alpha @ np.sin((2*np.pi/period)*t + phi) + error

    return FDataGrid(sample_points=t, data_matrix=y)


def make_multimodal_landmarks(n_samples: int=15, *, n_modes: int=1,
                              ndim_domain: int=1, ndim_image: int = 1,
                              start: float=-1, stop: float=1, std: float=.05,
                              random_state=None):
    """Generate landmarks points.

    Used by :func:`make_multimodal_samples` to generate the location of the
    landmarks.

    Generates a matrix containing the landmarks or locations of the modes of the
    samples generates by :func:`make_multimodal_samples`.

    If the same random state is used when generating the landmarks and
    multimodal samples, these will correspond to the position of the modes of
    the multimodal samples.

    Args:
        n_samples: Total number of samples.
        n_modes: Number of modes of each sample.
        ndim_domain: Number of dimensions of the domain.
        ndim_image: Number of dimensions of the image
        start: Starting point of the samples. In multidimensional objects the
            starting point of the axis.
        stop: Ending point of the samples. In multidimensional objects the
            ending point of the axis.
        std: Standard deviation of the variation of the modes location.
        random_state: Random state.

    Returns:
        :class:`np.ndarray` with the location of the modes, where the component
        (i,j,k) corresponds to the mode k of the image dimension j of the sample
        i.
    """

    random_state = sklearn.utils.check_random_state(random_state)

    modes_location = np.linspace(start, stop, n_modes + 2)[1:-1]
    modes_location = np.repeat(modes_location[:, np.newaxis], ndim_domain,
                               axis=1)


    variation = random_state.multivariate_normal((0,) * ndim_domain,
                                                 std * np.eye(ndim_domain),
                                                 size=(n_samples,
                                                       ndim_image,
                                                       n_modes))

    return modes_location + variation


def make_multimodal_samples(n_samples: int=15, *, n_modes: int=1,
                            points_per_dim: int=100, ndim_domain: int=1,
                            ndim_image: int=1, start: float=-1, stop: float=1.,
                            std: float=.05, mode_std: float=.02,
                            noise: float=.0, modes_location=None,
                            random_state=None):

    """Generate multimodal samples.

    Each sample :math:`x_i(t)` is proportional to a gaussian mixture, generated
    as the sum of multiple pdf of multivariate normal distributions with
    different means.

    .. math::

        x_i(t) \\propto \\sum_{n=1}^{\\text{n_modes}} \\exp \\left (
        {-\\frac{1}{2\\sigma} (t-\\mu_n)^T \\mathbb{1} (t-\\mu_n)} \\right )

    Where :math:`\\mu_n=\\text{mode_location}_n+\\epsilon` and :math:`\\epsilon`
    is normally distributed, with mean :math:`\\mathbb{0}` and standard
    deviation given by the parameter `std`.

    Args:
        n_samples: Total number of samples.
        n_modes: Number of modes of each sample.
        points_per_dim: Points per sample. If the object is multidimensional
            indicates the number of points for each dimension in the domain.
            The sample will have :math:
            `\\text{points_per_dim}^\\text{ndim_domain}` points of
            discretization.
        ndim_domain: Number of dimensions of the domain.
        ndim_image: Number of dimensions of the image
        start: Starting point of the samples. In multidimensional objects the
            starting point of each axis.
        stop: Ending point of the samples. In multidimensional objects the
            ending point of each axis.
        std: Standard deviation of the variation of the modes location.
        mode_std: Standard deviation :math:`\\sigma` of each mode.
        noise: Standard deviation of Gaussian noise added to the data.
        modes_location:  List of coordinates of each mode.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the samples.
    """

    random_state = sklearn.utils.check_random_state(random_state)


    if modes_location is None:

        location = make_multimodal_landmarks(n_samples=n_samples,
                                             n_modes=n_modes,
                                             ndim_domain=ndim_domain,
                                             ndim_image=ndim_image,
                                             start=start,
                                             stop=stop,
                                             std=std,
                                             random_state=random_state)

    else:
        location = np.asarray(modes_location)

        shape = (n_samples, ndim_image, n_modes, ndim_domain)
        location = location.reshape(shape)


    axis = np.linspace(start, stop, points_per_dim)

    if ndim_domain == 1:
        sample_points = axis
        evaluation_grid = axis
    else:
        sample_points = np.repeat(axis[:, np.newaxis], ndim_domain, axis=1).T

        meshgrid = np.meshgrid(*sample_points)

        evaluation_grid = np.empty(meshgrid[0].shape + (ndim_domain,))

        for i in range(ndim_domain):
            evaluation_grid[..., i] = meshgrid[i]

    # Data matrix of the grid
    shape = (n_samples,) + ndim_domain * (points_per_dim,) + (ndim_image,)
    data_matrix = np.zeros(shape)

    # Covariance matrix of the samples
    cov = mode_std * np.eye(ndim_domain)

    for i in range(n_samples):
        for j in range(ndim_image):
            for k in range(n_modes):
                data_matrix[i,...,j] += multivariate_normal.pdf(evaluation_grid,
                                                                mean=location[i,j,k],
                                                                cov=cov)

    # Constant to make modes value aprox. 1
    data_matrix *= (2*np.pi*mode_std)**(ndim_domain/2)

    data_matrix += random_state.normal(0, noise, size=data_matrix.shape)

    return FDataGrid(sample_points=sample_points, data_matrix=data_matrix)


def make_random_warping(n_samples: int=15, n_features: int=100, *,
                        start: float=0., stop: float=1., sigma: float=1.,
                        shape_parameter: float=50, n_random: int=4,
                        random_state=None):
    r"""Generate random warping functions.

     Let :math:`v(t)` be a randomly generated function defined in :math:`[0,1]`

     .. math::
        v(t) = \sum_{j=0}^{N} a_j \sin(\frac{2 \pi j}{K}t) + b_j
        \cos(\frac{2 \pi j}{K}t)

    where :math:`a_j, b_j \sim N(0, \sigma)`.

    The random warping it is constructed making an exponential map to
    :math:`\Gamma`.

     .. math::
        \gamma(t) = \int_0^t \left ( \frac{\sin(\|v\|)}{\|v\|} v(s) +
        \cos(\|v\|) \right )^2 ds

    An affine traslation it is used to define the warping in :math:`[a,b]`.

    The smoothing and shape of the warpings can be controlling changing
    :math:`N`, :math:`\sigma` and :math:`K= 1 + \text{shape_parameter}`.


    Args:
        n_samples: Total number of samples. Defaults 15.
        n_features: The total number of trajectories. Defaults 100.
        start: Starting point of the samples. Defaults 1.
        stop: Ending point of the samples. Defaults 0.
        sigma: Parameter to control the variance of the samples. Defaults 1.
        shape_parameter: Parameter to control the shape of the warpings.
            Should be a positive value. When the shape parameter goes to
            infinity the warpings generated are :math:`\gamma_{id}`.
            Defaults to 50.
        n_random: Number of random sines and cosines to be sum to construct
            the warpings.
        random_state: Random state.

    Returns:
        :class:`FDataGrid` object comprising all the samples.

     """
     # Based on the original implementation of J. D. Tucker in the
     # package python_fdasrsf <https://github.com/jdtuck/fdasrsf_python>.

    random_state = sklearn.utils.check_random_state(random_state)


    freq = shape_parameter + 1

    # Frequency
    omega = 2 * np.pi / freq
    sqrt2 = np.sqrt(2)
    sqrt_sigma = np.sqrt(sigma)

     # Originally it is compute in (0,1), then it is rescaled
    time = np.outer(np.linspace(0, 1, n_features), np.ones(n_samples))

    # Operates trasposed to broadcast dimensions
    v = np.outer(np.ones(n_features),
                 random_state.normal(scale=sqrt_sigma, size=n_samples))

    for j in range(2,2+n_random):
        alpha = random_state.normal(scale=sqrt_sigma, size=(2, n_samples))
        alpha *= sqrt2
        v += alpha[0] * np.cos(j*omega*time)
        v += alpha[1] * np.sin(j*omega*time)

    v -= v.mean(axis=0)

    # Equivalent to RN norm (equispaced monitorization points)
    v_norm = np.linalg.norm(v, axis=0)
    v_norm /= np.sqrt(n_features)

    # Exponential map of vectors
    v *= np.sin(v_norm) / v_norm
    v += np.cos(v_norm)
    np.square(v, out=v)

    # Creation of FDataGrid in the corresponding domain
    data_matrix = scipy.integrate.cumtrapz(v, dx=1./n_features, initial=0,
                                           axis=0)
    warping = FDataGrid(data_matrix.T, sample_points=time[:,0])
    warping = normalize_warping(warping, domain_range=(start, stop))
    warping.interpolator = SplineInterpolator(interpolation_order=3,
                                              monotone=True)


    return warping



"""Registration of functional data module.

This module contains methods to perform the registration of
functional data, in basis as well in discretized form.
"""

from ._landmark_registration import (landmark_shift_deltas,
                                     landmark_shift,
                                     landmark_registration_warping,
                                     landmark_registration)

from ._shift_registration import shift_registration, shift_registration_deltas

from ._registration_utils import (mse_decomposition,
                                  invert_warping,
                                  normalize_warping,
                                  _normalize_scale)

from ._elastic import (to_srsf, from_srsf,
                       elastic_registration,
                       elastic_registration_warping,
                       elastic_mean, warping_mean)



import numbers

import numpy as np


def _transform_to_2d(t):
    """Transform 1d arrays in column vectors."""
    t = np.asarray(t)

    dim = len(t.shape)
    assert dim <= 2

    if dim < 2:
        t = np.atleast_2d(t).T

    return t


def _execute_covariance(covariance, x, y):
    """Execute a covariance function.
    """
    x = _transform_to_2d(x)
    y = _transform_to_2d(y)

    if isinstance(covariance, numbers.Number):
        return covariance
    else:
        if callable(covariance):
            result = covariance(x, y)
        else:
            # GPy kernel
            result = covariance.K(x, y)

        assert result.shape[0] == len(x)
        assert result.shape[1] == len(y)
        return result


class Brownian():
    """Brownian covariance"""

    def __init__(self, *, variance: float=1., origin: float=0.):
        self.variance = variance
        self.origin = origin

    def __call__(self, x, y):
        """Brownian covariance function"""
        x = np.asarray(x) - self.origin
        y = np.asarray(y) - self.origin

        return self.variance * (np.abs(x) + np.abs(y.T) - np.abs(x - y.T)) / 2

    def _repr_latex_(self):
        return (r"\[K(x, y) = \sigma^2 \frac{|x - \mathcal{O}| + "
                r"|y - \mathcal{O}| - |x-y|}{2}\]"
                "where:"
                r"\begin{align*}"
                fr"\qquad\sigma^2 &= {self.variance} \\"
                fr"\mathcal{{O}} &= {self.origin} \\"
                r"\end{align*}")



import unittest

import numpy
import scipy.stats.mstats

from skfda.exploratory import stats
from skfda import FDataGrid


class TestFDataGrid(unittest.TestCase):

    # def setUp(self): could be defined for set up before any test

    def test_init(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        numpy.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            numpy.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]]))
        numpy.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        numpy.testing.assert_array_equal(
            fd.sample_points, numpy.array([[0., 0.25, 0.5, 0.75, 1.]]))

    def test_mean(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        mean = stats.mean(fd)
        numpy.testing.assert_array_equal(
            mean.data_matrix[0, ..., 0],
            numpy.array([1.5, 2.5, 3.5, 4.5, 5.5]))
        numpy.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        numpy.testing.assert_array_equal(
            fd.sample_points,
            numpy.array([[0., 0.25, 0.5, 0.75, 1.]]))

    def test_gmean(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        mean = stats.gmean(fd)
        numpy.testing.assert_array_equal(
            mean.data_matrix[0, ..., 0],
            scipy.stats.mstats.gmean(
                numpy.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])))
        numpy.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        numpy.testing.assert_array_equal(
            fd.sample_points,
            numpy.array([[0., 0.25, 0.5, 0.75, 1.]]))

    def test_slice(self):
        t = 10
        fd = FDataGrid(data_matrix=numpy.ones(t))
        fd = fd[:, 0]
        numpy.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            numpy.array([[1]]))
        numpy.testing.assert_array_equal(
            fd.sample_points,
            numpy.array([[0]]))


if __name__ == '__main__':
    print()
    unittest.main()



"""Functional principal component analysis.
"""

from ....exploratory.stats import mean
import numpy as np


def fpca(fdatagrid, n=2):
    """Compute Functional Principal Components Analysis.

    Performs Functional Principal Components Analysis to reduce
    dimensionality and obtain the principal modes of variation for a
    functional data object.

    It uses SVD numpy implementation to compute PCA.

    Args:
        fdatagrid (FDataGrid): functional data object.
        n (int, optional): Number of principal components. Defaults to 2.

    Returns:
        tuple: (scores, principal directions, eigenvalues)

    """
    fdatagrid = fdatagrid - mean(fdatagrid)  # centers the data
    # singular value decomposition
    u, s, v = np.linalg.svd(fdatagrid.data_matrix)
    principal_directions = v.T  # obtain the eigenvectors matrix
    eigenvalues = (np.diag(s) ** 2) / (fdatagrid.nsamples - 1)
    scores = u @ s  # functional principal scores

    return scores, principal_directions, eigenvalues




import scipy.integrate
import numpy


from ..representation import FData
from ..representation import FDataGrid
from ..preprocessing.registration import (
    normalize_warping, _normalize_scale, to_srsf,
    elastic_registration_warping)


def _cast_to_grid(fdata1, fdata2, eval_points=None):
    """Checks if the fdatas passed as argument are unidimensional and compatible
    and converts them to FDatagrid to compute their distances.


    Args:
        fdata1: (:obj:`FData`): First functional object.
        fdata2: (:obj:`FData`): Second functional object.

    Returns:
        tuple: Tuple with two :obj:`FDataGrid` with the same sample points.
    """

    # To allow use numpy arrays internally
    if (not isinstance(fdata1, FData) and not isinstance(fdata2, FData)
        and eval_points is not None):
        fdata1 = FDataGrid([fdata1], sample_points=eval_points)
        fdata2 = FDataGrid([fdata1], sample_points=eval_points)

    # Checks dimension
    elif (fdata2.ndim_image != fdata1.ndim_image or
          fdata2.ndim_domain != fdata1.ndim_domain):
        raise ValueError("Objects should have the same dimensions")

    # Case different domain ranges
    elif not numpy.array_equal(fdata1.domain_range, fdata2.domain_range):
        raise ValueError("Domain ranges for both objects must be equal")

    # Case new evaluation points specified
    elif eval_points is not None:
        if not numpy.array_equal(eval_points, fdata1.sample_points[0]):
            fdata1 = fdata1.to_grid(eval_points)
        if not numpy.array_equal(eval_points, fdata2.sample_points[0]):
            fdata2 = fdata2.to_grid(eval_points)

    elif not isinstance(fdata1, FDataGrid) and isinstance(fdata2, FDataGrid):
        fdata1 = fdata1.to_grid(fdata2.eval_points)

    elif not isinstance(fdata2, FDataGrid) and isinstance(fdata1, FDataGrid):
        fdata2 = fdata2.to_grid(fdata1.eval_points)

    elif not isinstance(fdata1, FDataGrid) and not isinstance(fdata2, FDataGrid):
        fdata1 = fdata1.to_grid(eval_points)
        fdata2 = fdata2.to_grid(eval_points)

    elif not numpy.array_equal(fdata1.sample_points,
                                      fdata2.sample_points):
        raise ValueError("Sample points for both objects must be equal or"
                         "a new list evaluation points must be specified")

    return fdata1, fdata2

def vectorial_norm(fdatagrid, p=2):
    r"""Apply a vectorial norm to a multivariate function.

    Given a multivariate function :math:`f:\mathbb{R}^n\rightarrow \mathbb{R}^d`
    applies a vectorial norm :math:`\| \cdot \|` to produce a function
    :math:`\|f\|:\mathbb{R}^n\rightarrow \mathbb{R}`.

    For example, let :math:`f:\mathbb{R} \rightarrow \mathbb{R}^2` be
    :math:`f(t)=(f_1(t), f_2(t))` and :math:`\| \cdot \|_2` the euclidian norm.

    .. math::
        \|f\|_2(t) = \sqrt { |f_1(t)|^2 + |f_2(t)|^2 }

    In general if :math:`p \neq \pm \infty` and :math:`f:\mathbb{R}^n
    \rightarrow \mathbb{R}^d`

    .. math::
        \|f\|_p(x_1, ... x_n) = \left ( \sum_{k=1}^{d} |f_k(x_1, ..., x_n)|^p
        \right )^{(1/p)}

    Args:
        fdatagrid (:class:`FDatagrid`): Functional object to be transformed.
        p (int, optional): Exponent in the lp norm. If p is a number then
            it is applied sum(abs(x)**p)**(1./p), if p is inf then max(abs(x)),
            and if p is -inf it is applied min(abs(x)). See numpy.linalg.norm
            to more information. Defaults to 2.

    Returns:
        (:class:`FDatagrid`): FDatagrid with image dimension equal to 1.

    Examples:

        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.misc.metrics import vectorial_norm

        First we will construct an example dataset with curves in
        :math:`\mathbb{R}^2`.

        >>> fd = make_multimodal_samples(ndim_image=2, random_state=0)
        >>> fd.ndim_image
        2

        We will apply the euclidean norm

        >>> fd = vectorial_norm(fd, p=2)
        >>> fd.ndim_image
        1

    """
    data_matrix = numpy.linalg.norm(fdatagrid.data_matrix, ord=p, axis=-1,
                                    keepdims=True)

    return fdatagrid.copy(data_matrix=data_matrix)


def distance_from_norm(norm, **kwargs):
    r"""Returns the distance induced by a norm.

    Given a norm :math:`\| \cdot \|: X \rightarrow \mathbb{R}`,
    returns the distance :math:`d: X \times X \rightarrow \mathbb{R}` induced by
    the norm:

    .. math::
        d(f,g) = \|f - g\|

    Args:
        norm (:obj:`Function`): Norm function `norm(fdata, **kwargs)`.
        **kwargs (dict, optional): Named parameters to be passed to the norm
            function.

    Returns:
        :obj:`Function`: Distance function `norm_distance(fdata1, fdata2)`.

    Examples:
        Computes the :math:`\mathbb{L}^2` distance between an object containing
        functional data corresponding to the function :math:`y(x) = x` defined
        over the interval [0, 1] and another one containing data of the function
        :math:`y(x) = x/2`.

        Firstly we create the functional data.

        >>> x = numpy.linspace(0, 1, 1001)
        >>> fd = FDataGrid([x], x)
        >>> fd2 =  FDataGrid([x/2], x)

        To construct the :math:`\mathbb{L}^2` distance it is used the
        :math:`\mathbb{L}^2` norm wich it is used to compute the distance.

        >>> l2_distance = distance_from_norm(norm_lp, p=2)
        >>> d = l2_distance(fd, fd2)
        >>> float('%.3f'% d)
        0.289

    """
    def norm_distance(fdata1, fdata2):
        # Substract operation checks if objects are compatible
        return norm(fdata1 - fdata2, **kwargs)

    norm_distance.__name__ = f"{norm.__name__}_distance"

    return norm_distance

def pairwise_distance(distance, **kwargs):
    r"""Return pairwise distance for FDataGrid objects.

    Given a distance returns the corresponding pairwise distance function.

    The pairwise distance calculates the distance between all possible pairs of
    one sample of the first FDataGrid object and one of the second one.

    The matrix returned by the pairwise distance is a matrix with as many rows
    as samples in the first object and as many columns as samples in the second
    one. Each element (i, j) of the matrix is the distance between the ith
    sample of the first object and the jth sample of the second one.

    Args:
        distance (:obj:`Function`): Distance functions between two functional
            objects `distance(fdata1, fdata2, **kwargs)`.
        **kwargs (:obj:`dict`, optional): parameters dictionary to be passed
            to the distance function.

    Returns:
        :obj:`Function`: Pairwise distance function, wich accepts two functional
        data objects and returns the pairwise distance matrix.
    """
    def pairwise(fdata1, fdata2):

        # Checks
        if not numpy.array_equal(fdata1.domain_range, fdata2.domain_range):
            raise ValueError("Domain ranges for both objects must be equal")

        # Creates an empty matrix with the desired size to store the results.
        matrix = numpy.empty((fdata1.nsamples, fdata2.nsamples))

        # Iterates over the different samples of both objects.
        for i in range(fdata1.nsamples):
            for j in range(fdata2.nsamples):
                matrix[i, j] = distance(fdata1[i], fdata2[j], **kwargs)
        # Computes the metric between all piars of x and y.
        return matrix

    pairwise.__name__ = f"pairwise_{distance.__name__}"

    return pairwise


def norm_lp(fdatagrid, p=2, p2=2):
    r"""Calculate the norm of all the samples in a FDataGrid object.

    For each sample sample f the Lp norm is defined as:

    .. math::
        \lVert f \rVert = \left( \int_D \lvert f \rvert^p dx \right)^{
        \frac{1}{p}}

    Where D is the domain over which the functions are defined.

    The integral is approximated using Simpson's rule.

    In general, if f is a multivariate function :math:`(f_1, ..., f_d)`, and
    :math:`D \subset \mathbb{R}^n`, it is applied the following generalization
    of the Lp norm.

    .. math::
        \| f \| = \left( \int_D \| f \|_{*}^p dx \right)^{
        \frac{1}{p}}

    Where :math:`\| \cdot \|_*` denotes a vectorial norm. See
    :func:`vectorial_norm` to more information.

    For example, if :math:`f: \mathbb{R}^2 \rightarrow \mathbb{R}^2`, and
    :math:`\| \cdot \|_*` is the euclidean norm
    :math:`\| (x,y) \|_* = \sqrt{x^2 + y^2}`, the lp norm applied is

    .. math::
        \lVert f \rVert = \left( \int \int_D \left ( \sqrt{ \lvert f_1(x,y)
        \rvert^2 + \lvert f_2(x,y) \rvert^2 } \right )^p dxdy \right)^{
        \frac{1}{p}}


    Args:
        fdatagrid (FDataGrid): FDataGrid object.
        p (int, optional): p of the lp norm. Must be greater or equal
            than 1. Defaults to 2.
        p2 (int, optional): p index of the vectorial norm applied in case of
            multivariate objects. Defaults to 2.

    Returns:
        numpy.darray: Matrix with as many rows as samples in the first
        object and as many columns as samples in the second one. Each
        element (i, j) of the matrix is the inner product of the ith sample
        of the first object and the jth sample of the second one.

    Examples:
        Calculates the norm of a FDataGrid containing the functions y = 1
        and y = x defined in the interval [0,1].


        >>> x = numpy.linspace(0,1,1001)
        >>> fd = FDataGrid([numpy.ones(len(x)), x] ,x)
        >>> norm_lp(fd).round(2)
        array([ 1.  ,  0.58])

        The lp norm is only defined if p >= 1.

        >>> norm_lp(fd, p = 0.5)
        Traceback (most recent call last):
            ....
        ValueError: p must be equal or greater than 1.

    """
    # Checks that the lp normed is well defined
    if p < 1:
        raise ValueError(f"p must be equal or greater than 1.")

    if fdatagrid.ndim_image > 1:
        data_matrix = numpy.linalg.norm(fdatagrid.data_matrix, ord=p2, axis=-1,
                                        keepdims=True)
    else:
        data_matrix = numpy.abs(fdatagrid.data_matrix)

    if fdatagrid.ndim_domain == 1:

        # Computes the norm, approximating the integral with Simpson's rule.
        res = scipy.integrate.simps(data_matrix[..., 0] ** p,
                                    x=fdatagrid.sample_points) ** (1 / p)

    else:
        # Needed to perform surface integration
        return NotImplemented

    if len(res) == 1:
        return res[0]

    return res


def lp_distance(fdata1, fdata2, p=2, *, eval_points=None):
    r"""Lp distance for FDataGrid objects.

    Calculates the distance between all possible pairs of one sample of
    the first FDataGrid object and one of the second one.

    For each pair of samples f and g the distance between them is defined as:

    .. math::
        d(f, g) = d(f, g) = \lVert f - g \rVert

    The norm is specified as a parameter but defaults to the l2 norm.

    Examples:
        Computes the distances between an object containing functional data
        corresponding to the functions y = 1 and y = x defined over the
        interval [0, 1] and another ones containing data of the functions y
        = 0 and y = x/2. The result then is an array 2x2 with the computed
        l2 distance between every pair of functions.

        >>> x = numpy.linspace(0, 1, 1001)
        >>> fd = FDataGrid([numpy.ones(len(x))], x)
        >>> fd2 =  FDataGrid([numpy.zeros(len(x))], x)
        >>> lp_distance(fd, fd2).round(2)
        1.0


        If the functional data are defined over a different set of points of
        discretisation the functions returns an exception.

        >>> x = numpy.linspace(0, 2, 1001)
        >>> fd2 =  FDataGrid([numpy.zeros(len(x)), x/2 + 0.5], x)
        >>> lp_distance(fd, fd2)
        Traceback (most recent call last):
            ....
        ValueError: Domain ranges for both objects must be equal

    """
    # Checks

    fdata1, fdata2 = _cast_to_grid(fdata1, fdata2, eval_points=eval_points)

    return norm_lp(fdata1 - fdata2, p=p)




def fisher_rao_distance(fdata1, fdata2, *, eval_points=None):
    """Compute the Fisher-Rao distance btween two functional objects.

    Let :math:`f_i` and :math:`f_j` be two functional observations, and let
    :math:`q_i` and :math:`q_j` be the corresponding SRSF (see :func:`to_srsf`),
    the fisher rao distance is defined as

    .. math::
        d_{FR}(f_i, f_j) = \\| q_i - q_j \\|_2 =
        \\left ( \\int_0^1 sgn(\\dot{f_i}(t))\\sqrt{|\\dot{f_i}(t)|} -
        sgn(\\dot{f_j}(t))\\sqrt{|\\dot{f_j}(t)|} dt \\right )^{\\frac{1}{2}}

    If the observations are distributions of random variables the distance will
    match with the usual fisher-rao distance in non-parametric form for
    probability distributions [S11-2]_.

    If the samples are defined in a domain different than (0,1) their domains
    are normalized to this interval with an affine transformation.

    Args:
        fdata1 (FData): First FData object.
        fdata2 (FData): Second FData object.
        eval_points (array_like, optional): Array with points of evaluation.

    Returns:
        Fisher rao distance.

    Raises:
        ValueError: If the objects are not unidimensional.

    Refereces:
        .. [S11-2] Srivastava, Anuj et. al. Registration of Functional Data
            Using Fisher-Rao Metric (2011). In *Function Representation and
            Metric* (pp. 5-7). arXiv:1103.3817v2.

    """

    fdata1, fdata2 = _cast_to_grid(fdata1, fdata2, eval_points=eval_points)

    # Both should have the same sample points
    eval_points_normalized = _normalize_scale(fdata1.sample_points[0])

    # Calculate the corresponding srsf and normalize to (0,1)
    fdata1 = fdata1.copy(sample_points=eval_points_normalized,
                         domain_range=(0,1))
    fdata2 = fdata2.copy(sample_points=eval_points_normalized,
                         domain_range=(0,1))

    fdata1_srsf = to_srsf(fdata1)
    fdata2_srsf = to_srsf(fdata2)

    # Return the L2 distance of the SRSF
    return lp_distance(fdata1_srsf, fdata2_srsf, p=2)

def amplitude_distance(fdata1, fdata2, *, lam=0., eval_points=None, **kwargs):
    """Compute the amplitude distance between two functional objects.

    Let :math:`f_i` and :math:`f_j` be two functional observations, and let
    :math:`q_i` and :math:`q_j` be the corresponding SRSF (see :func:`to_srsf`),
    the amplitude distance is defined as

    .. math::
        d_{A}(f_i, f_j)=min_{\\gamma \\in \\Gamma}d_{FR}(f_i \\circ \\gamma,f_j)

    A penalty term could be added to restrict the ammount of elasticity in the
    alignment used.

    .. math::
        d_{\\lambda}^2(f_i, f_j) =min_{\\gamma \\in \\Gamma} \\{
        d_{FR}^2(f_i \\circ \\gamma, f_j) + \\lambda \\mathcal{R}(\\gamma) \\}


    Where :math:`d_{FR}` is the Fisher-Rao distance and the penalty term is
    given by

    .. math::
        \\mathcal{R}(\\gamma) = \\|\\sqrt{\\dot{\\gamma}}- 1 \\|_{\\mathbb{L}^2}^2

    See [SK16-4-10-1]_ for a detailed explanation.

    If the samples are defined in a domain different than (0,1) their domains
    are normalized to this interval with an affine transformation.

    Args:
        fdata1 (FData): First FData object.
        fdata2 (FData): Second FData object.
        lam (float, optional): Penalty term to restric the elasticity.
        eval_points (array_like, optional): Array with points of evaluation.
        **kwargs (dict): Name arguments to be passed to
            :func:`elastic_registration_warping`.

    Returns:
        float: Elastic distance.

    Raises:
        ValueError: If the objects are not unidimensional.

    Refereces:
        ..  [SK16-4-10-1] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Amplitude Space and a Metric Structure*
            (pp. 107-109). Springer.
    """

    fdata1, fdata2 = _cast_to_grid(fdata1, fdata2, eval_points=eval_points)

    # Both should have the same sample points
    eval_points_normalized = _normalize_scale(fdata1.sample_points[0])

    # Calculate the corresponding srsf and normalize to (0,1)
    fdata1 = fdata1.copy(sample_points=eval_points_normalized,
                         domain_range=(0,1))
    fdata2 = fdata2.copy(sample_points=eval_points_normalized,
                         domain_range=(0,1))

    fdata1_srsf = to_srsf(fdata1)
    fdata2_srsf = to_srsf(fdata2)

    warping = elastic_registration_warping(fdata1,
                                             template=fdata2,
                                             lam=lam,
                                             eval_points=eval_points_normalized,
                                             fdatagrid_srsf=fdata1_srsf,
                                             template_srsf=fdata2_srsf,
                                             **kwargs)

    fdata1_reg = fdata1.compose(warping)

    distance = lp_distance(to_srsf(fdata1_reg), fdata2_srsf)

    if lam != 0.0:
        # L2 norm || sqrt(Dh) - 1 ||^2
        penalty = warping(eval_points_normalized, derivative=1,
                          keepdims=False)[0]
        penalty = numpy.sqrt(penalty, out=penalty)
        penalty -= 1
        penalty = numpy.square(penalty, out=penalty)
        penalty = scipy.integrate.simps(penalty, x=eval_points_normalized)

        distance = numpy.sqrt(distance**2 + lam*penalty)

    return distance

def phase_distance(fdata1, fdata2, *, lam=0., eval_points=None, **kwargs):
    """Compute the amplitude distance btween two functional objects.

    Let :math:`f_i` and :math:`f_j` be two functional observations, and let
    :math:`\\gamma_{ij}` the corresponding warping used in the elastic
    registration to align :math:`f_i` to :math:`f_j` (see
    :func:`elastic_registration`). The phase distance between :math:`f_i`
    and :math:`f_j` is defined as

    .. math::
        d_{P}(f_i, f_j) = d_{FR}(\\gamma_{ij}, \\gamma_{id}) =
        arcos \\left ( \\int_0^1 \\sqrt {\\dot \\gamma_{ij}(t)} dt \\right )

    See [SK16-4-10-2]_ for a detailed explanation.

    If the samples are defined in a domain different than (0,1) their domains
    are normalized to this interval with an affine transformation.

    Args:
        fdata1 (FData): First FData object.
        fdata2 (FData): Second FData object.
        lambda (float, optional): Penalty term to restric the elasticity.
        eval_points (array_like, optional): Array with points of evaluation.
        **kwargs (dict): Name arguments to be passed to
            :func:`elastic_registration_warping`.

    Returns:
        float: Phase distance between the objects.

    Raises:
        ValueError: If the objects are not unidimensional.


    Refereces:
        ..  [SK16-4-10-2] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Phase Space and a Metric Structure*
            (pp. 109-111). Springer.

    """

    fdata1, fdata2 = _cast_to_grid(fdata1, fdata2, eval_points=eval_points)

    # Rescale in (0,1)
    eval_points_normalized = _normalize_scale(fdata1.sample_points[0])

    # Calculate the corresponding srsf and normalize to (0,1)
    fdata1 = fdata1.copy(sample_points=eval_points_normalized,
                         domain_range=(0,1))
    fdata2 = fdata2.copy(sample_points=eval_points_normalized,
                         domain_range=(0,1))

    warping = elastic_registration_warping(fdata1, template=fdata2,
                                             lam=lam,
                                             eval_points=eval_points_normalized,
                                             **kwargs)

    derivative_warping = warping(eval_points_normalized, keepdims=False,
                                 derivative=1)[0]

    derivative_warping = numpy.sqrt(derivative_warping, out=derivative_warping)

    d = scipy.integrate.simps(derivative_warping, x=eval_points_normalized)

    return numpy.arccos(d)


def warping_distance(warping1, warping2, *, eval_points=None):
    """Compute the distance between warpings functions.

    Let :math:`\\gamma_i` and :math:`\\gamma_j` be two warpings, defined in
    :math:`\\gamma_i:[a,b] \\rightarrow [a,b]`. The distance in the
    space of warping functions, :math:`\\Gamma`, with the riemannian metric
    given by the fisher-rao inner product can be computed using the structure
    of hilbert sphere in their srsf's.

    .. math::
        d_{\\Gamma}(\\gamma_i, \\gamma_j) = cos^{-1} \\left ( \\int_0^1
        \\sqrt{\\dot \\gamma_i(t)\\dot \\gamma_j(t)}dt \\right )

    See [SK16-4-11-2]_ for a detailed explanation.

    If the warpings are not defined in [0,1], an affine transformation is maked
    to change the domain.

    Args:
        fdata1 (:obj:`FData`): First warping.
        fdata2 (:obj:`FData`): Second warping.
        eval_points (array_like, optional): Array with points of evaluation.

    Returns:
        float: Distance between warpings:

    Raises:
        ValueError: If the objects are not unidimensional.

    Refereces:
        ..  [SK16-4-11-2] Srivastava, Anuj & Klassen, Eric P. (2016). Functional
            and shape data analysis. In *Probability Density Functions*
            (pp. 113-117). Springer.

    """

    warping1, warping2 = _cast_to_grid(warping1, warping2,
                                       eval_points=eval_points)

    # Normalization of warping to (0,1)x(0,1)
    warping1 = normalize_warping(warping1, (0,1))
    warping2 = normalize_warping(warping2, (0,1))

    warping1_data = warping1.derivative().data_matrix[0, ..., 0]
    warping2_data = warping2.derivative().data_matrix[0, ..., 0]

    # In this case the srsf is the sqrt(gamma')
    srsf_warping1 = numpy.sqrt(warping1_data, out=warping1_data)
    srsf_warping2 = numpy.sqrt(warping2_data, out=warping2_data)

    product = numpy.multiply(srsf_warping1, srsf_warping2, out=srsf_warping1)
    d = scipy.integrate.simps(product, x=warping1.sample_points[0])

    return numpy.arccos(d)



"""Module with generic methods"""

import numpy as np


def _list_of_arrays(original_array):
    """Convert to a list of arrays.

    If the original list is one-dimensional (e.g. [1, 2, 3]), return list to
    array (in this case [array([1, 2, 3])]).

    If the original list is two-dimensional (e.g. [[1, 2, 3], [4, 5]]), return
    a list containing other one-dimensional arrays (in this case
    [array([1, 2, 3]), array([4, 5, 6])]).

    In any other case the behaviour is unespecified.

    """
    new_array = np.array([np.asarray(i) for i in
                          np.atleast_1d(original_array)])

    # Special case: Only one array, expand dimension
    if len(new_array.shape) == 1 and not any(isinstance(s, np.ndarray)
                                             for s in new_array):
        new_array = np.atleast_2d(new_array)

    return list(new_array)


def _coordinate_list(axes):
    """Convert a list with axes in a list with coordinates.

    Computes the cartesian product of the axes and returns a numpy array of
    1 dimension with all the possible combinations, for an arbitrary number of
    dimensions.

    Args:
        Axes (array_like): List with axes.

    Return:
        (np.ndarray): Numpy 2-D array with all the possible combinations.
        The entry (i,j) represent the j-th coordinate of the i-th point.

    Examples:

        >>> from skfda.representation._functional_data import _coordinate_list
        >>> axes = [[0,1],[2,3]]
        >>> _coordinate_list(axes)
        array([[0, 2],
               [0, 3],
               [1, 2],
               [1, 3]])

        >>> axes = [[0,1],[2,3],[4]]
        >>> _coordinate_list(axes)
        array([[0, 2, 4],
               [0, 3, 4],
               [1, 2, 4],
               [1, 3, 4]])

        >>> axes = [[0,1]]
        >>> _coordinate_list(axes)
        array([[0],
               [1]])

    """
    return np.vstack(list(map(np.ravel, np.meshgrid(*axes, indexing='ij')))).T



import unittest

import numpy as np

import matplotlib.pyplot as plt

from skfda import FDataGrid
from skfda.datasets import make_multimodal_samples
from skfda.misc.metrics import (fisher_rao_distance, amplitude_distance,
                                phase_distance, pairwise_distance, lp_distance,
                                warping_distance)
from skfda.preprocessing.registration import (
    elastic_registration, to_srsf, from_srsf,
    elastic_registration_warping, invert_warping, normalize_warping)

metric = pairwise_distance(lp_distance)
pairwise_fisher_rao = pairwise_distance(fisher_rao_distance)


class TestElasticRegistration(unittest.TestCase):
    """Test elastic registration"""


    def setUp(self):
        """Initialization of samples"""
        template = make_multimodal_samples(n_samples=1, std=0, random_state=1)
        self.template = template
        self.template_rep = template.concatenate(template).concatenate(template)
        self.unimodal_samples = make_multimodal_samples(n_samples=3,
                                                        random_state=1)

        t = np.linspace(-3, 3, 9)
        self.dummy_sample = FDataGrid([np.sin(t)], t)


    def test_to_srsf(self):
        """Test to srsf"""
        # Checks SRSF conversion
        srsf = to_srsf(self.dummy_sample)

        data_matrix = [[[-0.92155896], [-0.75559027], [ 0.25355399],
                        [ 0.81547327], [ 0.95333713], [ 0.81547327],
                        [ 0.25355399], [-0.75559027], [-0.92155896]]]

        np.testing.assert_almost_equal(data_matrix, srsf.data_matrix)


    def test_from_srsf(self):
        """Test from srsf"""

        # Checks SRSF conversion
        srsf = from_srsf(self.dummy_sample)

        data_matrix = [[[ 0.        ], [-0.23449228], [-0.83464009],
                        [-1.38200046], [-1.55623723], [-1.38200046],
                        [-0.83464009], [-0.23449228], [ 0.        ]]]

        np.testing.assert_almost_equal(data_matrix, srsf.data_matrix)


    def test_srsf_conversion(self):
        """Converts to srsf and pull backs"""
        initial = self.unimodal_samples(-1)
        converted = from_srsf(to_srsf(self.unimodal_samples), initial=initial)

        # Distances between original samples and s -> to_srsf -> from_srsf
        distances = np.diag(metric(converted, self.unimodal_samples))

        np.testing.assert_allclose(distances, 0, atol=8e-3)


    def test_template_alignment(self):
        """Test alignment to 1 template"""
        register = elastic_registration(self.unimodal_samples, self.template)
        distances = metric(self.template, register)

        np.testing.assert_allclose(distances, 0, atol=12e-3)

    def test_one_to_one_alignment(self):
        """Test alignment to 1 sample to a template"""
        register = elastic_registration(self.unimodal_samples[0], self.template)
        distances = metric(self.template, register)

        np.testing.assert_allclose(distances, 0, atol=12e-3)


    def test_set_alignment(self):
        """Test alignment 3 curves to set with 3 templates"""
        # Should give same result than test_template_alignment
        register = elastic_registration(self.unimodal_samples,
                                        self.template_rep)
        distances = metric(self.template, register)

        np.testing.assert_allclose(distances, 0, atol=12e-3)


    def test_simetry_of_aligment(self):
        """Check registration using inverse composition"""
        warping = elastic_registration_warping(self.unimodal_samples,
                                               self.template)
        inverse = invert_warping(warping)
        register = self.template_rep.compose(inverse)
        distances = np.diag(metric(self.unimodal_samples, register))

        np.testing.assert_allclose(distances, 0, atol=12e-3)

class TestElasticDistances(unittest.TestCase):
    """Test elastic distances"""

    def test_fisher_rao(self):
        """Test fisher rao distance"""

        t = np.linspace(0, 1, 100)
        sample  = FDataGrid([t, 1-t], t)
        f = np.square(sample)
        g = np.power(sample, 0.5)

        distance = [[0.62825868, 1.98009242], [1.98009242, 0.62825868]]
        res = pairwise_fisher_rao(f, g)

        np.testing.assert_almost_equal(res, distance, decimal=3)

    def test_fisher_rao_invariance(self):
        """Test invariance of fisher rao metric: d(f,g)= d(foh, goh)"""

        t = np.linspace(0, np.pi)
        id = FDataGrid([t], t)
        cos = np.cos(id)
        sin = np.sin(id)
        gamma = normalize_warping(np.sqrt(id), (0, np.pi))
        gamma2 = normalize_warping(np.square(id), (0, np.pi))

        distance_original = fisher_rao_distance(cos, sin)

        # Construction of 2 warpings
        distance_warping = fisher_rao_distance(cos.compose(gamma),
                                               sin.compose(gamma))
        distance_warping2 = fisher_rao_distance(cos.compose(gamma2),
                                               sin.compose(gamma2))

        # The error ~0.001 due to the derivation
        np.testing.assert_almost_equal(distance_original, distance_warping,
                                       decimal=2)

        np.testing.assert_almost_equal(distance_original, distance_warping2,
                                       decimal=2)

    def test_amplitude_distance_limit(self):
        """Test limit of amplitude distance penalty"""

        f = make_multimodal_samples(n_samples=1, random_state=1)
        g = make_multimodal_samples(n_samples=1, random_state=9999)

        amplitude_limit = amplitude_distance(f, g, lam=1000)
        fr_distance = fisher_rao_distance(f, g)

        np.testing.assert_almost_equal(amplitude_limit, fr_distance)


    def test_phase_distance_id(self):
        """Test of phase distance invariance"""
        f = make_multimodal_samples(n_samples=1, random_state=1)

        phase = phase_distance(f, 2*f)

        np.testing.assert_allclose(phase, 0, atol=1e-7)

    def test_warping_distance(self):
        """Test of warping distance"""
        t = np.linspace(0, 1)
        w1 = FDataGrid([t**5], t)
        w2 = FDataGrid([t**3], t)

        d = warping_distance(w1, w2)
        np.testing.assert_allclose(d,  np.arccos(np.sqrt(15)/4), atol=1e-3)

        d = warping_distance(w2, w2)
        np.testing.assert_allclose(d, 0, atol=2e-2)




if __name__ == '__main__':
    print()
    unittest.main()



"""Shift Registration of functional data module.

This module contains methods to perform the registration of
functional data using shifts, in basis as well in discretized form.
"""

import numpy
import scipy.integrate

__author__ = "Pablo Marcos Manchón"
__email__ = "pablo.marcosm@estudiante.uam.es"


def shift_registration_deltas(fd, *, maxiter=5, tol=1e-2, restrict_domain=False,
                              extrapolation=None, step_size=1, initial=None,
                              eval_points=None):
    r"""Return the lists of shifts used in the shift registration procedure.

        Realizes a registration of the curves, using shift aligment, as is
        defined in [RS05-7-2-1]_. Calculates :math:`\delta_{i}` for each sample
        such that :math:`x_i(t + \delta_{i})` minimizes the least squares
        criterion:

        .. math::
            \text{REGSSE} = \sum_{i=1}^{N} \int_{\mathcal{T}}
            [x_i(t + \delta_i) - \hat\mu(t)]^2 ds

        Estimates the shift parameter :math:`\delta_i` iteratively by
        using a modified Newton-Raphson algorithm, updating the mean
        in each iteration, as is described in detail in [RS05-7-9-1-1]_.

        Method only implemented for Funtional objects with domain and image
        dimension equal to 1.

    Args:
        fd (:class:`FData`): Functional data object to be registered.
        maxiter (int, optional): Maximun number of iterations.
            Defaults to 5.
        tol (float, optional): Tolerance allowable. The process will stop if
            :math:`\max_{i}|\delta_{i}^{(\nu)}-\delta_{i}^{(\nu-1)}|<tol`.
            Default sets to 1e-2.
        restrict_domain (bool, optional): If True restricts the domain to
            avoid evaluate points outside the domain using extrapolation.
            Defaults uses extrapolation.
        extrapolation (str or :class:`Extrapolation`, optional): Controls the
            extrapolation mode for elements outside the domain range.
            By default uses the method defined in fd. See :module:
            `extrapolation` to obtain more information.
        step_size (int or float, optional): Parameter to adjust the rate of
            convergence in the Newton-Raphson algorithm, see [RS05-7-9-1-1]_.
            Defaults to 1.
        initial (array_like, optional): Initial estimation of shifts.
            Default uses a list of zeros for the initial shifts.
        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain the discrete
            representation of the object to integrate. If None is
            passed it calls numpy.linspace in FDataBasis and uses the
            `sample_points` in FDataGrids.

    Returns:
        :class:`numpy.ndarray`: list with the shifts.

    Raises:
        ValueError: If the initial array has different length than the
            number of samples.

    Examples:

        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.basis import Fourier
        >>> from skfda.preprocessing.registration import shift_registration_deltas
        >>> fd = make_sinusoidal_process(n_samples=2, error_std=0, random_state=1)

        Registration of data in discretized form:

        >>> shift_registration_deltas(fd).round(3)
        array([-0.022,  0.03 ])

        Registration of data in basis form:

        >>> fd = fd.to_basis(Fourier())
        >>> shift_registration_deltas(fd).round(3)
        array([-0.022,  0.03 ])


    References:
        ..  [RS05-7-2-1] Ramsay, J., Silverman, B. W. (2005). Shift
            registration. In *Functional Data Analysis* (pp. 129-132).
            Springer.
        ..  [RS05-7-9-1-1] Ramsay, J., Silverman, B. W. (2005). Shift
            registration by the Newton-Raphson algorithm. In *Functional
            Data Analysis* (pp. 142-144). Springer.
    """

    # Initial estimation of the shifts

    if fd.ndim_image > 1 or fd.ndim_domain > 1:
        raise NotImplementedError("Method for unidimensional data.")

    domain_range = fd.domain_range[0]

    if initial is None:
        delta = numpy.zeros(fd.nsamples)

    elif len(initial) != fd.nsamples:
        raise ValueError(f"the initial shift ({len(initial)}) must have the "
                         f"same length than the number of samples "
                         f"({fd.nsamples})")
    else:
        delta = numpy.asarray(initial)

    # Fine equispaced mesh to evaluate the samples
    if eval_points is None:

        try:
            eval_points = fd.sample_points[0]
            nfine = len(eval_points)
        except AttributeError:
            nfine = max(fd.nbasis*10+1, 201)
            eval_points = numpy.linspace(*domain_range, nfine)

    else:
        nfine = len(eval_points)
        eval_points = numpy.asarray(eval_points)

    # Auxiliar arrays to avoid multiple memory allocations
    delta_aux = numpy.empty(fd.nsamples)
    tfine_aux = numpy.empty(nfine)

    # Computes the derivate of originals curves in the mesh points
    D1x = fd.evaluate(eval_points, derivative=1, keepdims=False)

    # Second term of the second derivate estimation of REGSSE. The
    # first term has been dropped to improve convergence (see references)
    d2_regsse = scipy.integrate.trapz(numpy.square(D1x), eval_points,
                                      axis=1)

    max_diff = tol + 1
    iter = 0

    # Auxiliar array if the domain will be restricted
    if restrict_domain:
        D1x_tmp = D1x
        tfine_tmp = eval_points
        tfine_aux_tmp = tfine_aux
        domain = numpy.empty(nfine, dtype=numpy.dtype(bool))

    ones = numpy.ones(fd.nsamples)
    eval_points_rep = numpy.outer(ones, eval_points)

    # Newton-Rhapson iteration
    while max_diff > tol and iter < maxiter:

        # Updates the limits for non periodic functions ignoring the ends
        if restrict_domain:
            # Calculates the new limits
            a = domain_range[0] - min(numpy.min(delta), 0)
            b = domain_range[1] - max(numpy.max(delta), 0)

            # New interval is (a,b)
            numpy.logical_and(tfine_tmp >= a, tfine_tmp <= b, out=domain)
            eval_points = tfine_tmp[domain]
            tfine_aux = tfine_aux_tmp[domain]
            D1x = D1x_tmp[:, domain]
            # Reescale the second derivate could be other approach
            # d2_regsse =
            #     d2_regsse_original * ( 1 + (a - b) / (domain[1] - domain[0]))
            d2_regsse = scipy.integrate.trapz(numpy.square(D1x),
                                              eval_points, axis=1)
            eval_points_rep = numpy.outer(ones, eval_points)

        # Computes the new values shifted
        x = fd.evaluate(eval_points_rep + numpy.atleast_2d(delta).T,
                        aligned_evaluation=False,
                        extrapolation=extrapolation,
                        keepdims=False)

        x.mean(axis=0, out=tfine_aux)

        # Calculates x - mean
        numpy.subtract(x, tfine_aux, out=x)

        d1_regsse = scipy.integrate.trapz(numpy.multiply(x, D1x, out=x),
                                          eval_points, axis=1)
        # Updates the shifts by the Newton-Rhapson iteration
        # delta = delta - step_size * d1_regsse / d2_regsse
        numpy.divide(d1_regsse, d2_regsse, out=delta_aux)
        numpy.multiply(delta_aux, step_size, out=delta_aux)
        numpy.subtract(delta, delta_aux, out=delta)

        # Updates convergence criterions
        max_diff = numpy.abs(delta_aux, out=delta_aux).max()
        iter += 1

    return delta


def shift_registration(fd, *, maxiter=5, tol=1e-2, restrict_domain=False,
                       extrapolation=None, step_size=1, initial=None,
                       eval_points=None, **kwargs):
    r"""Perform shift registration of the curves.

        Realizes a registration of the curves, using shift aligment, as is
        defined in [RS05-7-2]_. Calculates :math:`\delta_{i}` for each sample
        such that :math:`x_i(t + \delta_{i})` minimizes the least squares
        criterion:

        .. math::
            \text{REGSSE} = \sum_{i=1}^{N} \int_{\mathcal{T}}
            [x_i(t + \delta_i) - \hat\mu(t)]^2 ds

        Estimates the shift parameter :math:`\delta_i` iteratively by
        using a modified Newton-Raphson algorithm, updating the mean
        in each iteration, as is described in detail in [RS05-7-9-1]_.

    Args:
        fd (:class:`FData`): Functional data object to be registered.
        maxiter (int, optional): Maximun number of iterations.
            Defaults to 5.
        tol (float, optional): Tolerance allowable. The process will stop if
            :math:`\max_{i}|\delta_{i}^{(\nu)}-\delta_{i}^{(\nu-1)}|<tol`.
            Default sets to 1e-2.
        restrict_domain (bool, optional): If True restricts the domain to
            avoid evaluate points outside the domain using extrapolation.
            Defaults uses extrapolation.
        extrapolation (str or :class:`Extrapolation`, optional): Controls the
            extrapolation mode for elements outside the domain range.
            By default uses the method defined in fd. See :module:
            `extrapolation` to obtain more information.
        step_size (int or float, optional): Parameter to adjust the rate of
            convergence in the Newton-Raphson algorithm, see [RS05-7-9-1]_.
            Defaults to 1.
        initial (array_like, optional): Initial estimation of shifts.
            Default uses a list of zeros for the initial shifts.
        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain the discrete
            representation of the object to integrate. If None is
            passed it calls numpy.linspace in FDataBasis and uses the
            `sample_points` in FDataGrids.
        **kwargs: Keyword arguments to be passed to :func:`shift`.

    Returns:
        :class:`FData` A :class:`FData` object with the curves registered.

    Raises:
        ValueError: If the initial array has different length than the
            number of samples.

    Examples:

        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.basis import Fourier
        >>> from skfda.preprocessing.registration import shift_registration
        >>> fd = make_sinusoidal_process(n_samples=2, error_std=0, random_state=1)

        Registration of data in discretized form:

        >>> shift_registration(fd)
        FDataGrid(...)

        Registration of data in basis form:

        >>> fd = fd.to_basis(Fourier())
        >>> shift_registration(fd)
        FDataBasis(...)

    References:
        ..  [RS05-7-2] Ramsay, J., Silverman, B. W. (2005). Shift
            registration. In *Functional Data Analysis* (pp. 129-132).
            Springer.
        ..  [RS05-7-9-1] Ramsay, J., Silverman, B. W. (2005). Shift
            registration by the Newton-Raphson algorithm. In *Functional
            Data Analysis* (pp. 142-144). Springer.
    """

    delta = shift_registration_deltas(fd, maxiter=maxiter, tol=tol,
                                      restrict_domain=restrict_domain,
                                      extrapolation=extrapolation,
                                      step_size=step_size, initial=initial,
                                      eval_points=eval_points)

    # Computes the values with the final shift to construct the FDataBasis
    return fd.shift(delta, restrict_domain=restrict_domain,
                    extrapolation=extrapolation,
                    eval_points=eval_points, **kwargs)
